import numpy as np

from scvi import settings
from scvi.data import AnnDataManager
from scvi.dataloaders import DataSplitter
from scvi.dataloaders._data_splitting import (
    validate_data_split,
    validate_data_split_with_external_indexing,
)

from ._contrastive_dataloader import ContrastiveDataLoader


class ContrastiveDataSplitter(DataSplitter):
    """Creates ContrastiveDataLoader for training, validation, and test set.

    If ``train_size + validation_set < 1`` then ``test_set`` is non-empty.

    Parameters
    ----------
    adata_manager
        :class:`~scvi.data.AnnDataManager` object that has been created via
        ``setup_anndata``.
    background_indices: Indices for background samples in adata.
    target_indices: Indices for target samples in adata.
    train_size
        float, or None (default is 0.9)
    validation_size
        float, or None (default is None)
    shuffle_set_split
        Whether to shuffle indices before splitting. If `False`, the val, train, and
        test set are split in the sequential order of the data according to
        `validation_size` and `train_size` percentages.
    load_sparse_tensor
        If `True`, loads sparse CSR or CSC arrays in the input dataset as sparse
        :class:`~torch.Tensor` with the same layout. Can lead to significant
        speedups in transferring data to GPUs, depending on the sparsity of the data.
        Passed into :class:`~scvi.data.AnnDataLoader`.
    pin_memory
        Whether to copy tensors into device-pinned memory before returning them. Passed
        into :class:`~scvi.data.AnnDataLoader`.
    external_indexing
        A list of data split indices in the order of training, validation, and test sets.
        Validation and test set are not required and can be left empty.
        Note that per group (train,valid,test) it will cover both the background and target indices
    **kwargs
        Keyword args for data loader. Data loader class is
        :class:`~scvi.dataloaders.AnnDataLoader`.
    """

    def __init__(
        self,
        adata_manager: AnnDataManager,
        background_indices: list[int],
        target_indices: list[int],
        train_size: float | None = None,
        validation_size: float | None = None,
        shuffle_set_split: bool = True,
        load_sparse_tensor: bool = False,
        pin_memory: bool = False,
        external_indexing: list[np.array, np.array, np.array] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(
            adata_manager=adata_manager,
            train_size=train_size,
            validation_size=validation_size,
            shuffle_set_split=shuffle_set_split,
            load_sparse_tensor=load_sparse_tensor,
            pin_memory=pin_memory,
            **kwargs,
        )
        self.background_indices = background_indices
        self.target_indices = target_indices
        self.external_indexing = external_indexing

        self.n_background = len(background_indices)
        self.n_target = len(target_indices)
        if external_indexing is None:
            self.n_background_train, self.n_background_val = validate_data_split(
                self.n_background,
                self.train_size,
                self.validation_size,
                self.data_loader_kwargs.pop("batch_size", settings.batch_size),
                self.drop_last,
                self.train_size_is_none,
            )
            self.n_target_train, self.n_target_val = validate_data_split(
                self.n_target,
                self.train_size,
                self.validation_size,
                self.data_loader_kwargs.pop("batch_size", settings.batch_size),
                self.drop_last,
                self.train_size_is_none,
            )
        else:
            # we need to intersect the external indexing given with the bg/target indices
            self.background_train_idx, self.background_val_idx, self.background_test_idx = (
                np.intersect1d(self.external_indexing[n], self.background_indices)
                for n in range(3)
            )
            self.n_background_train, self.n_background_val = (
                validate_data_split_with_external_indexing(
                    self.n_background,
                    [self.background_train_idx, self.background_val_idx, self.background_test_idx],
                    self.data_loader_kwargs.pop("batch_size", settings.batch_size),
                    self.drop_last,
                )
            )
            self.background_train_idx, self.background_val_idx, self.background_test_idx = (
                self.background_train_idx.tolist(),
                self.background_val_idx.tolist(),
                self.background_test_idx.tolist(),
            )

            self.target_train_idx, self.target_val_idx, self.target_test_idx = (
                np.intersect1d(self.external_indexing[n], self.target_indices) for n in range(3)
            )
            self.n_target_train, self.n_target_val = validate_data_split_with_external_indexing(
                self.n_target,
                [self.target_train_idx, self.target_val_idx, self.target_test_idx],
                self.data_loader_kwargs.pop("batch_size", settings.batch_size),
                self.drop_last,
            )
            self.target_train_idx, self.target_val_idx, self.target_test_idx = (
                self.target_train_idx.tolist(),
                self.target_val_idx.tolist(),
                self.target_test_idx.tolist(),
            )

        self.n_train = self.n_background_train + self.n_target_train
        self.n_val = self.n_background_val + self.n_target_val

    def setup(self, stage: str | None = None):
        """Split background and target indices into train/val/test sets."""
        background_indices = self.background_indices
        n_background_train = self.n_background_train
        n_background_val = self.n_background_val

        target_indices = self.target_indices
        n_target_train = self.n_target_train
        n_target_val = self.n_target_val

        # Need to separate to the external and non-external cases of the target/bg indices
        if self.external_indexing is None:
            if self.shuffle_set_split:
                random_state = np.random.RandomState(seed=settings.seed)
                background_indices = random_state.permutation(background_indices).tolist()
                target_indices = random_state.permutation(target_indices).tolist()

            self.background_val_idx = background_indices[:n_background_val]
            self.background_train_idx = background_indices[
                n_background_val : (n_background_val + n_background_train)
            ]
            self.background_test_idx = background_indices[
                (n_background_val + n_background_train) :
            ]

            self.target_val_idx = target_indices[:n_target_val]
            self.target_train_idx = target_indices[n_target_val : (n_target_val + n_target_train)]
            self.target_test_idx = target_indices[(n_target_val + n_target_train) :]

        self.val_idx = self.background_val_idx + self.target_val_idx
        self.train_idx = self.background_train_idx + self.target_train_idx
        self.test_idx = self.background_test_idx + self.target_test_idx

    def train_dataloader(self) -> ContrastiveDataLoader:
        return ContrastiveDataLoader(
            adata_manager=self.adata_manager,
            background_indices=self.background_train_idx,
            target_indices=self.target_train_idx,
            shuffle=True,
            drop_last=self.drop_last,
            load_sparse_tensor=self.load_sparse_tensor,
            pin_memory=self.pin_memory,
            **self.data_loader_kwargs,
        )

    def val_dataloader(self) -> ContrastiveDataLoader:
        if len(self.background_val_idx) > 0 and len(self.target_val_idx) > 0:
            return ContrastiveDataLoader(
                adata_manager=self.adata_manager,
                background_indices=self.background_val_idx,
                target_indices=self.target_val_idx,
                shuffle=False,
                drop_last=False,
                load_sparse_tensor=self.load_sparse_tensor,
                pin_memory=self.pin_memory,
                **self.data_loader_kwargs,
            )
        else:
            pass

    def test_dataloader(self) -> ContrastiveDataLoader:
        if len(self.background_test_idx) > 0 and len(self.target_test_idx) > 0:
            return ContrastiveDataLoader(
                adata_manager=self.adata_manager,
                background_indices=self.background_test_idx,
                target_indices=self.target_test_idx,
                shuffle=False,
                drop_last=False,
                load_sparse_tensor=self.load_sparse_tensor,
                pin_memory=self.pin_memory,
                **self.data_loader_kwargs,
            )
        else:
            pass



from __future__ import annotations

from typing import Literal

import anndata
import jax.numpy as jnp
import mudata
import torch

Number = int | float
AnnOrMuData = anndata.AnnData | mudata.MuData
Tensor = torch.Tensor | jnp.ndarray
LossRecord = dict[str, Tensor] | Tensor
# TODO(adamgayoso): Add constants for minified data types.
MinifiedDataType = Literal["latent_posterior_parameters"]



from scvi.data import synthetic_iid
from scvi.external.velovi import VELOVI


def test_velovi():
    n_latent = 5
    adata = synthetic_iid()
    adata.layers["spliced"] = adata.X.copy()
    adata.layers["unspliced"] = adata.X.copy()
    VELOVI.setup_anndata(adata, unspliced_layer="unspliced", spliced_layer="spliced")
    model = VELOVI(adata, n_latent=n_latent)
    model.train(1, check_val_every_n_epoch=1, train_size=0.5)
    model.get_latent_representation()
    model.get_velocity()
    model.get_latent_time()
    model.get_state_assignment()
    model.get_expression_fit()
    model.get_directional_uncertainty()
    model.get_permutation_scores(labels_key="labels")

    _ = model.history

    # tests __repr__
    print(model)



from __future__ import annotations

import logging
import os
from pathlib import Path
from typing import TYPE_CHECKING

import torch
from lightning.pytorch import seed_everything
from rich.console import Console
from rich.logging import RichHandler

if TYPE_CHECKING:
    from typing import Literal

scvi_logger = logging.getLogger("scvi")


class ScviConfig:
    """Config manager for scvi-tools.

    Examples
    --------
    To set the seed

    >>> scvi.settings.seed = 1

    To set the batch size for functions like `SCVI.get_latent_representation`

    >>> scvi.settings.batch_size = 1024

    To set the progress bar style, choose one of "rich", "tqdm"

    >>> scvi.settings.progress_bar_style = "rich"

    To set the verbosity

    >>> import logging
    >>> scvi.settings.verbosity = logging.INFO

    To set the number of threads PyTorch will use

    >>> scvi.settings.num_threads = 2

    To prevent Jax from preallocating GPU memory on start (default)

    >>> scvi.settings.jax_preallocate_gpu_memory = False
    """

    def __init__(
        self,
        verbosity: int = logging.INFO,
        progress_bar_style: Literal["rich", "tqdm"] = "tqdm",
        batch_size: int = 128,
        seed: int | None = None,
        logging_dir: str = "./scvi_log/",
        dl_num_workers: int = 0,
        dl_persistent_workers: bool = False,
        jax_preallocate_gpu_memory: bool = False,
        warnings_stacklevel: int = 2,
    ):
        self.warnings_stacklevel = warnings_stacklevel
        self.seed = seed
        self.batch_size = batch_size
        if progress_bar_style not in ["rich", "tqdm"]:
            raise ValueError("Progress bar style must be in ['rich', 'tqdm']")
        self.progress_bar_style = progress_bar_style
        self.logging_dir = logging_dir
        self.dl_num_workers = dl_num_workers
        self.dl_persistent_workers = dl_persistent_workers
        self._num_threads = None
        self.jax_preallocate_gpu_memory = jax_preallocate_gpu_memory
        self.verbosity = verbosity

    @property
    def batch_size(self) -> int:
        """Minibatch size for loading data into the model.

        This is only used after a model is trained. Trainers have specific
        `batch_size` parameters.
        """
        return self._batch_size

    @batch_size.setter
    def batch_size(self, batch_size: int):
        """Minibatch size for loading data into the model.

        This is only used after a model is trained. Trainers have specific
        `batch_size` parameters.
        """
        self._batch_size = batch_size

    @property
    def dl_num_workers(self) -> int:
        """Number of workers for PyTorch data loaders (Default is 0)."""
        return self._dl_num_workers

    @dl_num_workers.setter
    def dl_num_workers(self, dl_num_workers: int):
        """Number of workers for PyTorch data loaders (Default is 0)."""
        self._dl_num_workers = dl_num_workers

    @property
    def dl_persistent_workers(self) -> bool:
        """Whether to use persistent_workers in PyTorch data loaders (Default is False)."""
        return self._dl_persistent_workers

    @dl_persistent_workers.setter
    def dl_persistent_workers(self, dl_persistent_workers: bool):
        """Whether to use persistent_workers in PyTorch data loaders (Default is False)."""
        self._dl_persistent_workers = dl_persistent_workers

    @property
    def logging_dir(self) -> Path:
        """Directory for training logs (default `'./scvi_log/'`)."""
        return self._logging_dir

    @logging_dir.setter
    def logging_dir(self, logging_dir: str | Path):
        self._logging_dir = Path(logging_dir).resolve()

    @property
    def num_threads(self) -> None:
        """Number of threads PyTorch will use."""
        return self._num_threads

    @num_threads.setter
    def num_threads(self, num: int):
        """Number of threads PyTorch will use."""
        self._num_threads = num
        torch.set_num_threads(num)

    @property
    def progress_bar_style(self) -> str:
        """Library to use for progress bar."""
        return self._pbar_style

    @progress_bar_style.setter
    def progress_bar_style(self, pbar_style: Literal["tqdm", "rich"]):
        """Library to use for progress bar."""
        self._pbar_style = pbar_style

    @property
    def seed(self) -> int:
        """Random seed for torch and numpy."""
        return self._seed

    @seed.setter
    def seed(self, seed: int | None = None):
        """Random seed for torch and numpy."""
        if seed is None:
            self._seed = None
        else:
            torch.backends.cudnn.deterministic = True
            torch.backends.cudnn.benchmark = False
            # Ensure deterministic CUDA operations for Jax (see https://github.com/google/jax/issues/13672)
            if "XLA_FLAGS" not in os.environ:
                os.environ["XLA_FLAGS"] = "--xla_gpu_deterministic_ops=true"
            else:
                os.environ["XLA_FLAGS"] += " --xla_gpu_deterministic_ops=true"
            seed_everything(seed)
            self._seed = seed

    @property
    def verbosity(self) -> int:
        """Verbosity level (default `logging.INFO`)."""
        return self._verbosity

    @verbosity.setter
    def verbosity(self, level: str | int):
        """Sets logging configuration for scvi based on chosen level of verbosity.

        If "scvi" logger has no StreamHandler, add one.
        Else, set its level to `level`.

        Parameters
        ----------
        level
            Sets "scvi" logging level to `level`
        force_terminal
            Rich logging option, set to False if piping to file output.
        """
        self._verbosity = level
        scvi_logger.setLevel(level)
        if len(scvi_logger.handlers) == 0:
            console = Console(force_terminal=True)
            if console.is_jupyter is True:
                console.is_jupyter = False
            ch = RichHandler(level=level, show_path=False, console=console, show_time=False)
            formatter = logging.Formatter("%(message)s")
            ch.setFormatter(formatter)
            scvi_logger.addHandler(ch)
        else:
            scvi_logger.setLevel(level)

    @property
    def warnings_stacklevel(self) -> int:
        """Stacklevel for warnings."""
        return self._warnings_stacklevel

    @warnings_stacklevel.setter
    def warnings_stacklevel(self, stacklevel: int):
        """Stacklevel for warnings."""
        self._warnings_stacklevel = stacklevel

    def reset_logging_handler(self):
        """Resets "scvi" log handler to a basic RichHandler().

        This is useful if piping outputs to a file.
        """
        scvi_logger.removeHandler(scvi_logger.handlers[0])
        ch = RichHandler(level=self._verbosity, show_path=False, show_time=False)
        formatter = logging.Formatter("%(message)s")
        ch.setFormatter(formatter)
        scvi_logger.addHandler(ch)

    @property
    def jax_preallocate_gpu_memory(self):
        """Jax GPU memory allocation settings.

        If False, Jax will ony preallocate GPU memory it needs.
        If float in (0, 1), Jax will preallocate GPU memory to that
        fraction of the GPU memory.
        """
        return self._jax_gpu

    @jax_preallocate_gpu_memory.setter
    def jax_preallocate_gpu_memory(self, value: float | bool):
        # see https://jax.readthedocs.io/en/latest/gpu_memory_allocation.html#gpu-memory-allocation
        if value is False:
            os.environ["XLA_PYTHON_CLIENT_PREALLOCATE"] = "false"
        elif isinstance(value, float):
            if value >= 1 or value <= 0:
                raise ValueError("Need to use a value between 0 and 1")
            # format is ".XX"
            os.environ["XLA_PYTHON_CLIENT_MEM_FRACTION"] = str(value)[1:4]
        else:
            raise ValueError("value not understood, need bool or float in (0, 1)")
        self._jax_gpu = value


settings = ScviConfig()



import logging

from lightning.pytorch.callbacks import ProgressBar as ProgressBarBase

from scvi.utils import track

logger = logging.getLogger(__name__)


class ProgressBar(ProgressBarBase):
    """Custom progress bar for scvi-tools models.

    Parameters
    ----------
    refresh_rate
        Determines at which rate (in number of epochs) the progress bars get updated.
        Set it to ``0`` to disable the display.
    """

    def __init__(self, refresh_rate: int = 1):
        super().__init__()
        if refresh_rate > 1:
            raise ValueError(
                "scvi-tools progress bar only supports a value of 0 of 1 for "
                "`progress_bar_refresh_rate`"
            )
        self._refresh_rate = refresh_rate
        self._enabled = True

    def __getstate__(self):
        # can't pickle the tqdm objects
        state = self.__dict__.copy()
        state["main_progress_bar"] = None
        return state

    @property
    def is_enabled(self) -> bool:
        return self._enabled and self.refresh_rate > 0

    @property
    def refresh_rate(self) -> int:
        return self._refresh_rate

    @property
    def is_disabled(self) -> bool:
        return not self.is_enabled

    def disable(self) -> None:
        self._enabled = False

    def enable(self) -> None:
        self._enabled = True

    def init_train_tqdm(self, trainer):
        """Override this to customize the tqdm bar for training."""
        bar = track(
            None,
            total=trainer.max_epochs,
            description="Training",
            style="tqdm",
            disable=self.is_disabled,
        )
        return bar

    def on_sanity_check_start(self, trainer, pl_module):
        super().on_sanity_check_start(trainer, pl_module)
        logger.info("Running sanity check on val set...")

    def on_train_start(self, trainer, pl_module):
        super().on_train_start(trainer, pl_module)
        self.main_progress_bar = self.init_train_tqdm(trainer)

    def on_train_epoch_start(self, trainer, pl_module):
        super().on_train_epoch_start(trainer, pl_module)
        if self._should_update(self.trainer.current_epoch, self.trainer.max_epochs):
            epoch = trainer.current_epoch + 1
            self.main_progress_bar.set_description(f"Epoch {epoch}/{trainer.max_epochs}")

    def _should_update(self, current, total):
        return self.is_enabled and (current % self.refresh_rate == 0 or current == total)

    def on_train_epoch_end(self, trainer, pl_module):
        super().on_train_epoch_end(trainer, pl_module)
        if self._should_update(self.trainer.current_epoch, self.trainer.max_epochs):
            self.main_progress_bar.update()
            self.main_progress_bar.set_postfix(self.get_metrics(trainer, pl_module))

    def on_train_end(self, trainer, pl_module):
        super().on_train_end(trainer, pl_module)
        if self.is_enabled:
            self.main_progress_bar.close()


def convert_inf(x):
    """The tqdm doesn't support inf values. We have to convert it to None."""
    if x == float("inf"):
        return None
    return x



from collections.abc import Callable

import rich
from anndata import AnnData
from mudata import MuData

from scvi._types import AnnOrMuData

from ._base_field import AnnDataField, BaseAnnDataField


class BaseMuDataWrapperClass(BaseAnnDataField):
    """A wrapper class that adds MuData support for an AnnDataField.

    The wrapper class makes calls to the ``self.adata_field`` instance using the appropriate
    modality and allows for manipulation at the MuData object level.

    Parameters
    ----------
    mod_key
        Modality key where data is stored. If ``None``, uses the top level MuData object
        attributes.
    mod_required
        If ``True``, raises ``ValueError`` when ``mod_key`` is ``None``.
    """

    def __init__(self, mod_key: str | None = None, mod_required: bool = False) -> None:
        super().__init__()
        if mod_required and mod_key is None:
            raise ValueError(f"Modality required for {self.__class__.__name__} but not provided.")
        self._mod_key = mod_key
        self._preregister = lambda _self, _mdata: None

    @property
    def adata_field(self) -> AnnDataField:
        """:class"`~scvi.data.fields.AnnDataField` instance that this class instance wraps."""
        return self._adata_field

    @property
    def registry_key(self) -> str:
        """The key that is referenced by models via a data loader."""
        return self.adata_field.registry_key

    @property
    def mod_key(self) -> str | None:
        """The modality key of the data field within the MuData (if applicable)."""
        return self._mod_key

    @property
    def attr_name(self) -> str:
        """The name of the AnnData/MuData attribute where the data is stored."""
        return self.adata_field.attr_name

    @property
    def attr_key(self) -> str | None:
        """The key of the data field within the relevant AnnData/MuData attribute."""
        return self.adata_field.attr_key

    @property
    def is_empty(self) -> bool:
        return self.adata_field.is_empty

    def get_modality(self, mdata: MuData) -> AnnOrMuData:
        """Fetches the appropriate modality from the MuData object using ``self.mod_key``."""
        if isinstance(mdata, AnnData):
            raise AssertionError("`get_modality` can only be called on MuData objects.")
        bdata = mdata
        if self.mod_key is not None:
            if self.mod_key not in mdata.mod:
                raise ValueError(f"Modality {self.mod_key} not found in mdata.mod.")
            bdata = mdata.mod[self.mod_key]
        return bdata

    def validate_field(self, mdata: MuData) -> None:
        """Validate the field."""
        if isinstance(mdata, AnnData):
            raise ValueError("`get_modality` can only be called on MuData objects.")
        bdata = self.get_modality(mdata)
        return self.adata_field.validate_field(bdata)

    def preregister(self, mdata: MuData) -> None:
        """Function that is called prior to registering fields.

        Function that is be called at the beginning of
        :meth:`~scvi.data.fields.BaseMuDataWrapperClass.register_field` and
        :meth:`~scvi.data.fields.BaseMuDataWrapperClass.transfer_field`. Used when data
        manipulation is necessary across modalities.
        """
        return self._preregister(self, mdata)

    def register_field(self, mdata: MuData) -> dict:
        """Register the field."""
        self.preregister(mdata)
        bdata = self.get_modality(mdata)
        return self.adata_field.register_field(bdata)

    def transfer_field(self, state_registry: dict, mdata_target: MuData, **kwargs) -> dict:
        """Transfer the field."""
        self.preregister(mdata_target)
        bdata_target = self.get_modality(mdata_target)
        return self.adata_field.transfer_field(state_registry, bdata_target, **kwargs)

    def get_summary_stats(self, state_registry: dict) -> dict:
        """Get summary stats."""
        return self.adata_field.get_summary_stats(state_registry)

    def view_state_registry(self, state_registry: dict) -> rich.table.Table | None:
        """View the state registry."""
        return self.adata_field.view_state_registry(state_registry)


def MuDataWrapper(
    adata_field_cls: AnnDataField, preregister_fn: Callable | None = None
) -> AnnDataField:
    """Wraps an AnnDataField with :class:`~scvi.data.fields.BaseMuDataWrapperClass`.

    Parameters
    ----------
    adata_field_cls
        AnnDataField class to wrap.
    preregister_fn
        Function that will be called at the beginning of
        :meth:`~scvi.data.fields.BaseMuDataWrapperClass.register_field` and
        :meth:`~scvi.data.fields.BaseMuDataWrapperClass.transfer_field`.
    """
    if not isinstance(adata_field_cls, type):
        raise ValueError("`adata_field_cls` must be a class, not an instance.")

    def mudata_field_init(
        self, *args, mod_key: str | None = None, mod_required: bool = False, **kwargs
    ):
        BaseMuDataWrapperClass.__init__(self, mod_key=mod_key, mod_required=mod_required)
        self._adata_field = adata_field_cls(*args, **kwargs)
        if preregister_fn is not None:
            self._preregister = preregister_fn

    return type(
        f"MuData{adata_field_cls.__name__}",
        (BaseMuDataWrapperClass,),
        {
            "__init__": mudata_field_init,
        },
    )



from __future__ import annotations

from typing import TYPE_CHECKING

from torch.distributions import Normal as NormalTorch

if TYPE_CHECKING:
    import torch


class Normal(NormalTorch):
    """Normal distribution.

    Parameters
    ----------
    mu
        loc of the Normal distribution.
    var
        scale of the Normal distribution.
    validate_args
        whether to validate input.
    normal_mu
        Normalized mean expression of the distribution.
        This optional parameter is not used in any computations, but allows to store
        normalization expression levels.

    """

    def __init__(
        self,
        mu: torch.Tensor,
        scale: torch.Tensor,
        validate_args: bool | None = None,
        normal_mu: torch.Tensor = None,
    ):
        super().__init__(loc=mu, scale=scale, validate_args=validate_args)
        self.normal_mu = normal_mu

    def __repr__(self) -> str:
        param_names = [k for k, _ in self.arg_constraints.items() if k in self.__dict__]
        args_string = ", ".join(
            [
                f"{p}: "
                f"{self.__dict__[p] if self.__dict__[p].numel() == 1 else self.__dict__[p].size()}"
                for p in param_names
                if self.__dict__[p] is not None
            ]
        )
        return self.__class__.__name__ + "(" + args_string + ")"

    def get_normalized(self, key) -> torch.Tensor:
        if key == "mu":
            return self.loc
        elif key == "scale":
            return self.normal_mu
        else:
            raise ValueError(f"normalized key {key} not recognized")



from typing import NamedTuple


class _SAVE_KEYS_NT(NamedTuple):
    ADATA_FNAME: str = "adata.h5ad"
    MDATA_FNAME: str = "mdata.h5mu"
    MODEL_FNAME: str = "model.pt"
    MODEL_STATE_DICT_KEY: str = "model_state_dict"
    VAR_NAMES_KEY: str = "var_names"
    ATTR_DICT_KEY: str = "attr_dict"
    LEGACY_MODEL_FNAME: str = "model_params.pt"
    LEGACY_VAR_NAMES_FNAME: str = "var_names.csv"
    LEGACY_SETUP_DICT_FNAME: str = "attr.pkl"


SAVE_KEYS = _SAVE_KEYS_NT()



from __future__ import annotations

from inspect import signature
from typing import TYPE_CHECKING

import torch

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator
    from typing import Any

    from torch import Tensor

    from scvi.module.base import LossOutput


def compute_elbo(
    module: Callable[[dict[str, Tensor | None], dict], tuple[Any, Any, LossOutput]],
    dataloader: Iterator[dict[str, Tensor | None]],
    return_mean: bool = True,
    **kwargs,
) -> float:
    """Compute the evidence lower bound (ELBO) on the data.

    The ELBO is the reconstruction error plus the Kullback-Leibler (KL) divergences between the
    variational distributions and the priors. It is different from the marginal log-likelihood;
    specifically, it is a lower bound on the marginal log-likelihood plus a term that is constant
    with respect to the variational distribution. It still gives good insights on the modeling of
    the data and is fast to compute.

    Parameters
    ----------
    module
        A callable (can be a :class:`~torch.nn.Module` instance) that takes a dictionary of
        :class:`~torch.Tensor`s as input and returns a tuple of three elements, where the last
        element is an instance of :class:`~scvi.module.base.LossOutput`.
    dataloader
        An iterator over minibatches of data on which to compute the metric. The minibatches
        should be formatted as a dictionary of :class:`~torch.Tensor` with keys as expected by
        the ``forward`` method of ``module``.
    **kwargs
        Additional keyword arguments to pass into ``module``.
    return_mean
        If ``True``, return the mean ELBO across the dataset. If ``False``, return the ELBO for
        each cell individually.

    Returns
    -------
    The evidence lower bound (ELBO) of the data.
    """
    elbo = []
    if "full_forward_pass" in signature(module._get_inference_input).parameters:
        get_inference_input_kwargs = {"full_forward_pass": True}
    else:
        get_inference_input_kwargs = {}
    for tensors in dataloader:
        _, _, losses = module(
            tensors, **kwargs, get_inference_input_kwargs=get_inference_input_kwargs
        )
        if isinstance(losses.reconstruction_loss, dict):
            reconstruction_loss = torch.stack(list(losses.reconstruction_loss.values())).sum(dim=0)
        else:
            reconstruction_loss = losses.reconstruction_loss
        if isinstance(losses.kl_local, dict):
            kl_local = torch.stack(list(losses.kl_local.values())).sum(dim=0)
        else:
            kl_local = losses.kl_local
        elbo.append(reconstruction_loss + kl_local)

    elbo = torch.cat(elbo, dim=0)
    if return_mean:
        elbo = elbo.mean()
    return elbo


def compute_reconstruction_error(
    module: Callable[[dict[str, Tensor | None], dict], tuple[Any, Any, LossOutput]],
    dataloader: Iterator[dict[str, Tensor | None]],
    return_mean: bool = True,
    **kwargs,
) -> dict[str, float]:
    """Compute the reconstruction error on the data.

    The reconstruction error is the negative log likelihood of the data given the latent
    variables. It is different from the marginal log-likelihood, but still gives good insights on
    the modeling of the data and is fast to compute.

    Parameters
    ----------
    module
        A callable (can be a :class:`~torch.nn.Module` instance) that takes a dictionary of
        :class:`~torch.Tensor`s as input and returns a tuple of three elements, where the last
        element is an instance of :class:`~scvi.module.base.LossOutput`.
    dataloader
        An iterator over minibatches of data on which to compute the metric. The minibatches
        should be formatted as a dictionary of :class:`~torch.Tensor` with keys as expected by
        the ``forward`` method of ``module``.
    return_mean
        If ``True``, return the mean reconstruction error across the dataset. If ``False``,
        return the reconstruction error for each cell individually.
    **kwargs
        Additional keyword arguments to pass into ``module``.

    Returns
    -------
    A dictionary of the reconstruction error of the data.
    """
    # Iterate once over the data and computes the reconstruction error
    if "full_forward_pass" in signature(module._get_inference_input).parameters:
        get_inference_input_kwargs = {"full_forward_pass": True}
    else:
        get_inference_input_kwargs = {}
    log_lkl = {}
    for tensors in dataloader:
        _, _, loss_output = module(
            tensors,
            loss_kwargs={"kl_weight": 1},
            get_inference_input_kwargs=get_inference_input_kwargs,
            **kwargs,
        )
        if not isinstance(loss_output.reconstruction_loss, dict):
            rec_loss_dict = {"reconstruction_loss": loss_output.reconstruction_loss}
        else:
            rec_loss_dict = loss_output.reconstruction_loss
        for key, value in rec_loss_dict.items():
            if key in log_lkl:
                log_lkl[key].append(value)
            else:
                log_lkl[key] = [value]

    for key, _ in log_lkl.items():
        log_lkl[key] = torch.cat(log_lkl[key], dim=0)
        if return_mean:
            log_lkl[key] = torch.mean(log_lkl[key])

    return log_lkl



from unittest import mock

import numpy as np
import pytest
from flax import linen as nn

from scvi.data import synthetic_iid
from scvi.model import JaxSCVI
from scvi.utils import attrdict


def test_jax_scvi(n_latent=5):
    adata = synthetic_iid()
    JaxSCVI.setup_anndata(
        adata,
        batch_key="batch",
    )
    model = JaxSCVI(adata, n_latent=n_latent)
    model.train(2, train_size=0.5, check_val_every_n_epoch=1)
    model.get_latent_representation()

    model = JaxSCVI(adata, n_latent=n_latent, gene_likelihood="poisson")
    model.train(1, train_size=0.5)
    z1 = model.get_latent_representation(give_mean=True, n_samples=1)
    assert z1.ndim == 2
    z2 = model.get_latent_representation(give_mean=False, n_samples=15)
    assert z2.ndim == 3
    assert z2.shape[0] == 15


def test_jax_scvi_training(n_latent: int = 5, dropout_rate: float = 0.1):
    adata = synthetic_iid()
    JaxSCVI.setup_anndata(
        adata,
        batch_key="batch",
    )

    model = JaxSCVI(adata, n_latent=n_latent, dropout_rate=dropout_rate)
    assert model.module.training

    with mock.patch("scvi.module._jaxvae.nn.Dropout", wraps=nn.Dropout) as mock_dropout_cls:
        mock_dropout = mock.Mock()
        mock_dropout.side_effect = lambda h, **_kwargs: h
        mock_dropout_cls.return_value = mock_dropout
        model.train(1, train_size=0.5, check_val_every_n_epoch=1)

        assert not model.module.training
        mock_dropout_cls.assert_called()
        mock_dropout.assert_has_calls(
            12 * [mock.call(mock.ANY, deterministic=False)]
            + 8 * [mock.call(mock.ANY, deterministic=True)]
        )


def test_jax_scvi_save_load(save_path: str, n_latent: int = 5):
    adata = synthetic_iid()
    JaxSCVI.setup_anndata(
        adata,
        batch_key="batch",
    )
    model = JaxSCVI(adata, n_latent=n_latent)
    model.train(2, train_size=0.5, check_val_every_n_epoch=1)
    z1 = model.get_latent_representation(adata)
    model.save(save_path, overwrite=True, save_anndata=True)
    model.view_setup_args(save_path)
    model = JaxSCVI.load(save_path)
    model.get_latent_representation()

    # Load with mismatched genes.
    tmp_adata = synthetic_iid(
        n_genes=200,
    )
    with pytest.raises(ValueError):
        JaxSCVI.load(save_path, adata=tmp_adata)

    # Load with different batches.
    tmp_adata = synthetic_iid()
    tmp_adata.obs["batch"] = tmp_adata.obs["batch"].cat.rename_categories(["batch_2", "batch_3"])
    with pytest.raises(ValueError):
        JaxSCVI.load(save_path, adata=tmp_adata)

    model = JaxSCVI.load(save_path, adata=adata)
    assert "batch" in model.adata_manager.data_registry
    assert model.adata_manager.data_registry.batch == attrdict(
        {"attr_name": "obs", "attr_key": "_scvi_batch"}
    )
    assert model.is_trained is True

    z2 = model.get_latent_representation()
    np.testing.assert_array_equal(z1, z2)



from __future__ import annotations

from typing import TYPE_CHECKING

from scipy.sparse import csr_matrix

from scvi import REGISTRY_KEYS

if TYPE_CHECKING:
    from anndata import AnnData
    from mudata import MuData

    from scvi.data import AnnDataManager


def get_minified_adata_scrna(
    adata_manager: AnnDataManager,
    keep_count_data: bool = False,
) -> AnnData:
    """Returns a minified AnnData.

    Parameters
    ----------
    adata_manager
        Manager with original AnnData, of which we want to create a minified version.
    keep_count_data
        If True, the count data is kept in the minified data. If False, the count data is removed.
    """
    adata = adata_manager.adata.copy()
    if keep_count_data:
        pass
    else:
        del adata.raw
        counts = adata_manager.get_from_registry(REGISTRY_KEYS.X_KEY)
        all_zeros = csr_matrix(counts.shape)
        X = all_zeros
        layers = {layer: all_zeros.copy() for layer in adata_manager.adata.layers}
        adata.X = X
        adata.layers = layers
    return adata


def get_minified_mudata(
    adata_manager: AnnDataManager,
    keep_count_data: bool = False,
) -> MuData:
    """Returns a minified MuData that works for most multi modality models (MULTIVI, TOTALVI).

    Parameters
    ----------
    adata_manager
        Manager with original MuData, of which we want to create a minified version.
    keep_count_data
        If True, the count data is kept in the minified data. If False, the count data is removed.
    """
    mdata = adata_manager.adata.copy()
    if keep_count_data:
        pass
    else:
        for modality in mdata.mod_names:
            del mdata[modality].raw
            all_zeros = csr_matrix(mdata[modality].X.shape)
            mdata[modality].X = all_zeros.copy()
            if len(mdata[modality].layers) > 0:
                layers = {layer: all_zeros.copy() for layer in mdata[modality].layers}
                mdata[modality].layers = layers
    return mdata



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import torch

from scvi.utils import unsupported_if_adata_minified

if TYPE_CHECKING:
    from collections.abc import Iterator, Sequence

    import numpy.typing as npt
    from anndata import AnnData
    from torch import Tensor
    from torch.distributions import Distribution


logger = logging.getLogger(__name__)


class VAEMixin:
    """Universal variational auto-encoder (VAE) methods."""

    @torch.inference_mode()
    @unsupported_if_adata_minified
    def get_elbo(
        self,
        adata: AnnData | None = None,
        indices: Sequence[int] | None = None,
        batch_size: int | None = None,
        dataloader: Iterator[dict[str, Tensor | None]] | None = None,
        return_mean: bool = True,
        **kwargs,
    ) -> float:
        """Compute the evidence lower bound (ELBO) on the data.

        The ELBO is the reconstruction error plus the Kullback-Leibler (KL) divergences between the
        variational distributions and the priors. It is different from the marginal log-likelihood;
        specifically, it is a lower bound on the marginal log-likelihood plus a term that is
        constant with respect to the variational distribution. It still gives good insights on the
        modeling of the data and is fast to compute.

        Parameters
        ----------
        adata
            :class:`~anndata.AnnData` object with :attr:`~anndata.AnnData.var_names` in the same
            order as the ones used to train the model. If ``None`` and ``dataloader`` is also
            ``None``, it defaults to the object used to initialize the model.
        indices
            Indices of observations in ``adata`` to use. If ``None``, defaults to all observations.
            Ignored if ``dataloader`` is not ``None``.
        batch_size
            Minibatch size for the forward pass. If ``None``, defaults to
            ``scvi.settings.batch_size``. Ignored if ``dataloader`` is not ``None``.
        dataloader
            An iterator over minibatches of data on which to compute the metric. The minibatches
            should be formatted as a dictionary of :class:`~torch.Tensor` with keys as expected by
            the model. If ``None``, a dataloader is created from ``adata``.
        return_mean
            Whether to return the mean of the ELBO or the ELBO for each observation.
        **kwargs
            Additional keyword arguments to pass into the forward method of the module.

        Returns
        -------
        Evidence lower bound (ELBO) of the data.

        Notes
        -----
        This is not the negative ELBO, so higher is better.
        """
        from scvi.model.base._log_likelihood import compute_elbo

        if adata is not None and dataloader is not None:
            raise ValueError("Only one of `adata` or `dataloader` can be provided.")
        elif dataloader is None:
            adata = self._validate_anndata(adata)
            dataloader = self._make_data_loader(
                adata=adata, indices=indices, batch_size=batch_size
            )

        return -compute_elbo(self.module, dataloader, return_mean=return_mean, **kwargs)

    @torch.inference_mode()
    @unsupported_if_adata_minified
    def get_marginal_ll(
        self,
        adata: AnnData | None = None,
        indices: Sequence[int] | None = None,
        n_mc_samples: int = 1_000,
        batch_size: int | None = None,
        return_mean: bool = True,
        dataloader: Iterator[dict[str, Tensor | None]] = None,
        **kwargs,
    ) -> float | Tensor:
        """Compute the marginal log-likehood of the data.

        The computation here is a biased estimator of the marginal log-likelihood of the data.

        Parameters
        ----------
        adata
            :class:`~anndata.AnnData` object with :attr:`~anndata.AnnData.var_names` in the same
            order as the ones used to train the model. If ``None`` and ``dataloader`` is also
            ``None``, it defaults to the object used to initialize the model.
        indices
            Indices of observations in ``adata`` to use. If ``None``, defaults to all observations.
            Ignored if ``dataloader`` is not ``None``.
        n_mc_samples
            Number of Monte Carlo samples to use for the estimator. Passed into the module's
            ``marginal_ll`` method.
        batch_size
            Minibatch size for the forward pass. If ``None``, defaults to
            ``scvi.settings.batch_size``. Ignored if ``dataloader`` is not ``None``.
        return_mean
            Whether to return the mean of the marginal log-likelihood or the marginal-log
            likelihood for each observation.
        dataloader
            An iterator over minibatches of data on which to compute the metric. The minibatches
            should be formatted as a dictionary of :class:`~torch.Tensor` with keys as expected by
            the model. If ``None``, a dataloader is created from ``adata``.
        **kwargs
            Additional keyword arguments to pass into the module's ``marginal_ll`` method.

        Returns
        -------
        If ``True``, returns the mean marginal log-likelihood. Otherwise returns a tensor of shape
        ``(n_obs,)`` with the marginal log-likelihood for each observation.

        Notes
        -----
        This is not the negative log-likelihood, so higher is better.
        """
        from numpy import mean

        if not hasattr(self.module, "marginal_ll"):
            raise NotImplementedError(
                "The model's module must implement `marginal_ll` to compute the marginal "
                "log-likelihood."
            )
        elif adata is not None and dataloader is not None:
            raise ValueError("Only one of `adata` or `dataloader` can be provided.")

        if dataloader is None:
            adata = self._validate_anndata(adata)
            dataloader = self._make_data_loader(
                adata=adata, indices=indices, batch_size=batch_size
            )

        log_likelihoods: list[float | Tensor] = [
            self.module.marginal_ll(
                tensors, n_mc_samples=n_mc_samples, return_mean=return_mean, **kwargs
            )
            for tensors in dataloader
        ]

        if return_mean:
            return mean(log_likelihoods)
        else:
            return torch.cat(log_likelihoods, dim=0)

    @torch.inference_mode()
    @unsupported_if_adata_minified
    def get_reconstruction_error(
        self,
        adata: AnnData | None = None,
        indices: Sequence[int] | None = None,
        batch_size: int | None = None,
        dataloader: Iterator[dict[str, Tensor | None]] | None = None,
        return_mean: bool = True,
        **kwargs,
    ) -> dict[str, float]:
        r"""Compute the reconstruction error on the data.

        The reconstruction error is the negative log likelihood of the data given the latent
        variables. It is different from the marginal log-likelihood, but still gives good insights
        on the modeling of the data and is fast to compute. This is typically written as
        :math:`p(x \mid z)`, the likelihood term given one posterior sample.

        Parameters
        ----------
        adata
            :class:`~anndata.AnnData` object with :attr:`~anndata.AnnData.var_names` in the same
            order as the ones used to train the model. If ``None`` and ``dataloader`` is also
            ``None``, it defaults to the object used to initialize the model.
        indices
            Indices of observations in ``adata`` to use. If ``None``, defaults to all observations.
            Ignored if ``dataloader`` is not ``None``
        batch_size
            Minibatch size for the forward pass. If ``None``, defaults to
            ``scvi.settings.batch_size``. Ignored if ``dataloader`` is not ``None``
        dataloader
            An iterator over minibatches of data on which to compute the metric. The minibatches
            should be formatted as a dictionary of :class:`~torch.Tensor` with keys as expected by
            the model. If ``None``, a dataloader is created from ``adata``.
        return_mean
            Whether to return the mean reconstruction loss or the reconstruction loss
            for each observation.
        **kwargs
            Additional keyword arguments to pass into the forward method of the module.

        Returns
        -------
        Reconstruction error for the data.

        Notes
        -----
        This is not the negative reconstruction error, so higher is better.
        """
        from scvi.model.base._log_likelihood import compute_reconstruction_error

        if adata is not None and dataloader is not None:
            raise ValueError("Only one of `adata` or `dataloader` can be provided.")

        if dataloader is None:
            adata = self._validate_anndata(adata)
            dataloader = self._make_data_loader(
                adata=adata, indices=indices, batch_size=batch_size
            )

        return compute_reconstruction_error(
            self.module, dataloader, return_mean=return_mean, **kwargs
        )

    @torch.inference_mode()
    def get_latent_representation(
        self,
        adata: AnnData | None = None,
        indices: Sequence[int] | None = None,
        give_mean: bool = True,
        mc_samples: int = 5_000,
        batch_size: int | None = None,
        return_dist: bool = False,
        dataloader: Iterator[dict[str, Tensor | None]] = None,
    ) -> npt.NDArray | tuple[npt.NDArray, npt.NDArray]:
        """Compute the latent representation of the data.

        This is typically denoted as :math:`z_n`.

        Parameters
        ----------
        adata
            :class:`~anndata.AnnData` object with :attr:`~anndata.AnnData.var_names` in the same
            order as the ones used to train the model. If ``None`` and ``dataloader`` is also
            ``None``, it defaults to the object used to initialize the model.
        indices
            Indices of observations in ``adata`` to use. If ``None``, defaults to all observations.
            Ignored if ``dataloader`` is not ``None``
        give_mean
            If ``True``, returns the mean of the latent distribution. If ``False``, returns an
            estimate of the mean using ``mc_samples`` Monte Carlo samples.
        mc_samples
            Number of Monte Carlo samples to use for the estimator for distributions with no
            closed-form mean (e.g., the logistic normal distribution). Not used if ``give_mean`` is
            ``True`` or if ``return_dist`` is ``True``.
        batch_size
            Minibatch size for the forward pass. If ``None``, defaults to
            ``scvi.settings.batch_size``. Ignored if ``dataloader`` is not ``None``
        return_dist
            If ``True``, returns the mean and variance of the latent distribution. Otherwise,
            returns the mean of the latent distribution.
        dataloader
            An iterator over minibatches of data on which to compute the metric. The minibatches
            should be formatted as a dictionary of :class:`~torch.Tensor` with keys as expected by
            the model. If ``None``, a dataloader is created from ``adata``.

        Returns
        -------
        An array of shape ``(n_obs, n_latent)`` if ``return_dist`` is ``False``. Otherwise, returns
        a tuple of arrays ``(n_obs, n_latent)`` with the mean and variance of the latent
        distribution.
        """
        from torch.distributions import Normal
        from torch.nn.functional import softmax

        from scvi.module._constants import MODULE_KEYS

        self._check_if_trained(warn=False)
        if adata is not None and dataloader is not None:
            raise ValueError("Only one of `adata` or `dataloader` can be provided.")

        if dataloader is None:
            adata = self._validate_anndata(adata)
            dataloader = self._make_data_loader(
                adata=adata, indices=indices, batch_size=batch_size
            )

        zs: list[Tensor] = []
        qz_means: list[Tensor] = []
        qz_vars: list[Tensor] = []
        for tensors in dataloader:
            outputs: dict[str, Tensor | Distribution | None] = self.module.inference(
                **self.module._get_inference_input(tensors)
            )

            if MODULE_KEYS.QZ_KEY in outputs:
                qz: Distribution = outputs.get(MODULE_KEYS.QZ_KEY)
                qzm: Tensor = qz.loc
                qzv: Tensor = qz.scale.square()
            else:
                qzm: Tensor = outputs.get(MODULE_KEYS.QZM_KEY)
                qzv: Tensor = outputs.get(MODULE_KEYS.QZV_KEY)
                qz: Distribution = Normal(qzm, qzv.sqrt())

            if return_dist:
                qz_means.append(qzm.cpu())
                qz_vars.append(qzv.cpu())
                continue

            z: Tensor = qzm if give_mean else outputs.get(MODULE_KEYS.Z_KEY)

            if give_mean and getattr(self.module, "latent_distribution", None) == "ln":
                samples = qz.sample([mc_samples])
                z = softmax(samples, dim=-1).mean(dim=0)

            zs.append(z.cpu())

        if return_dist:
            return torch.cat(qz_means).numpy(), torch.cat(qz_vars).numpy()
        else:
            return torch.cat(zs).numpy()



from anndata import AnnData
from mudata import MuData

from scvi import REGISTRY_KEYS
from scvi.data import AnnDataManager
from scvi.data._constants import _MODEL_NAME_KEY, _SETUP_ARGS_KEY
from scvi.data.fields import (
    CategoricalJointObsField,
    CategoricalObsField,
    LabelsWithUnlabeledObsField,
    LayerField,
    MuDataCategoricalJointObsField,
    MuDataCategoricalObsField,
    MuDataLayerField,
    MuDataNumericalJointObsField,
    MuDataProteinLayerField,
    NumericalJointObsField,
    ProteinObsmField,
)
from scvi.model import SCVI


def unsupervised_training_one_epoch(
    adata: AnnData,
    run_setup_anndata: bool = True,
    batch_key: str | None = None,
    labels_key: str | None = None,
):
    if run_setup_anndata:
        SCVI.setup_anndata(adata, batch_key=batch_key, labels_key=labels_key)
    m = SCVI(adata)
    m.train(1, train_size=0.4)


def generic_setup_adata_manager(
    adata: AnnData,
    batch_key: str | None = None,
    labels_key: str | None = None,
    categorical_covariate_keys: list[str] | None = None,
    continuous_covariate_keys: list[str] | None = None,
    layer: str | None = None,
    protein_expression_obsm_key: str | None = None,
    protein_names_uns_key: str | None = None,
) -> AnnDataManager:
    setup_args = locals()
    setup_args.pop("adata")
    setup_method_args = {_MODEL_NAME_KEY: "TestModel", _SETUP_ARGS_KEY: setup_args}

    batch_field = CategoricalObsField(REGISTRY_KEYS.BATCH_KEY, batch_key)
    anndata_fields = [
        batch_field,
        LayerField(REGISTRY_KEYS.X_KEY, layer, is_count_data=True),
        CategoricalObsField(REGISTRY_KEYS.LABELS_KEY, labels_key),
        CategoricalJointObsField(REGISTRY_KEYS.CAT_COVS_KEY, categorical_covariate_keys),
        NumericalJointObsField(REGISTRY_KEYS.CONT_COVS_KEY, continuous_covariate_keys),
    ]
    if protein_expression_obsm_key is not None:
        anndata_fields.append(
            ProteinObsmField(
                REGISTRY_KEYS.PROTEIN_EXP_KEY,
                protein_expression_obsm_key,
                use_batch_mask=True,
                batch_field=batch_field,
                colnames_uns_key=protein_names_uns_key,
                is_count_data=True,
            )
        )
    adata_manager = AnnDataManager(fields=anndata_fields, setup_method_args=setup_method_args)
    adata_manager.register_fields(adata)
    return adata_manager


def scanvi_setup_adata_manager(
    adata: AnnData,
    unlabeled_category: str,
    batch_key: str | None = None,
    labels_key: str | None = None,
) -> AnnDataManager:
    setup_args = locals()
    setup_args.pop("adata")
    setup_method_args = {_MODEL_NAME_KEY: "TestModel", _SETUP_ARGS_KEY: setup_args}
    anndata_fields = [
        CategoricalObsField(REGISTRY_KEYS.BATCH_KEY, batch_key),
        LabelsWithUnlabeledObsField(REGISTRY_KEYS.LABELS_KEY, labels_key, unlabeled_category),
    ]
    adata_manager = AnnDataManager(fields=anndata_fields, setup_method_args=setup_method_args)
    adata_manager.register_fields(adata)
    return adata_manager


def generic_setup_mudata_manager(
    mdata: MuData,
    layer_mod,
    layer: str | None = None,
    batch_mod: str | None = None,
    batch_key: str | None = None,
    categorical_covariate_mod: str | None = None,
    categorical_covariate_keys: list[str] | None = None,
    continuous_covariate_mod: str | None = None,
    continuous_covariate_keys: list[str] | None = None,
    protein_expression_mod: str | None = None,
    protein_expression_layer: str | None = None,
) -> AnnDataManager:
    setup_args = locals()
    setup_args.pop("mdata")
    setup_method_args = {_MODEL_NAME_KEY: "TestModel", _SETUP_ARGS_KEY: setup_args}

    batch_field = MuDataCategoricalObsField(REGISTRY_KEYS.BATCH_KEY, batch_key, mod_key=batch_mod)
    anndata_fields = [
        MuDataLayerField(
            REGISTRY_KEYS.X_KEY,
            layer,
            mod_key=layer_mod,
            is_count_data=True,
            mod_required=True,
        ),
        batch_field,
        MuDataCategoricalJointObsField(
            REGISTRY_KEYS.CAT_COVS_KEY,
            categorical_covariate_keys,
            mod_key=categorical_covariate_mod,
        ),
        MuDataNumericalJointObsField(
            REGISTRY_KEYS.CONT_COVS_KEY,
            continuous_covariate_keys,
            mod_key=continuous_covariate_mod,
        ),
    ]
    if protein_expression_mod is not None:
        anndata_fields.append(
            MuDataProteinLayerField(
                REGISTRY_KEYS.PROTEIN_EXP_KEY,
                protein_expression_layer,
                mod_key=protein_expression_mod,
                mod_required=True,
                use_batch_mask=True,
                batch_field=batch_field,
            )
        )
    mdata_manager = AnnDataManager(fields=anndata_fields, setup_method_args=setup_method_args)
    mdata_manager.register_fields(mdata)
    return mdata_manager



import json
import os
from dataclasses import asdict

import anndata
import numpy as np
import pytest

import scvi
from scvi.criticism import create_criticism_report
from scvi.data import synthetic_iid
from scvi.hub import HubMetadata, HubModel, HubModelCardHelper
from scvi.hub._constants import _SCVI_HUB


def prep_model() -> scvi.model.SCVI:
    adata = scvi.data.synthetic_iid()
    scvi.model.SCVI.setup_anndata(adata)
    model = scvi.model.SCVI(adata)
    model.train(1)
    return model


def prep_scvi_hub_model(save_path: str) -> HubModel:
    model = prep_model()
    model_path = os.path.join(save_path, "test_scvi")
    model.save(model_path, save_anndata=True, overwrite=True)
    create_criticism_report(model, save_folder=model_path, n_samples=2)

    metadata = HubMetadata.from_dir(model_path, anndata_version=anndata.__version__)
    desc = "scVI model trained on synthetic IID data and uploaded with the full training data."
    card = HubModelCardHelper.from_dir(
        model_path,
        license_info="cc-by-4.0",
        anndata_version=anndata.__version__,
        data_modalities=["rna"],
        data_is_annotated=False,
        description=desc,
    )
    return HubModel(model_path, metadata=metadata, model_card=card)


def prep_scvi_no_anndata_hub_model(save_path: str) -> HubModel:
    model = prep_model()
    model_path = os.path.join(save_path, "test_scvi_no_anndata")
    model.save(model_path, save_anndata=False, overwrite=True)

    metadata = HubMetadata.from_dir(model_path, anndata_version=anndata.__version__)
    card = HubModelCardHelper.from_dir(
        model_path,
        license_info="cc-by-4.0",
        anndata_version=anndata.__version__,
        data_modalities=["rna"],
        data_is_annotated=False,
        description="scVI model trained on synthetic IID data and uploaded with no data.",
    )
    return HubModel(model_path, metadata=metadata, model_card=card)


def prep_scvi_minified_hub_model(save_path: str) -> HubModel:
    model = prep_model()
    model_path = os.path.join(save_path, "test_scvi_minified")
    model.save(model_path, save_anndata=False, overwrite=True)

    qzm, qzv = model.get_latent_representation(return_dist=True, give_mean=False)
    model.adata.obsm["X_latent_qzm"] = qzm
    model.adata.obsm["X_latent_qzv"] = qzv
    model.minify_adata()
    model_path = os.path.join(save_path, "test_scvi_minified")
    model.save(model_path, save_anndata=True, overwrite=True)

    metadata = HubMetadata.from_dir(model_path, anndata_version=anndata.__version__)
    desc = "scVI model trained on synthetic IID data and uploaded with the minified data."
    card = HubModelCardHelper.from_dir(
        model_path,
        license_info="cc-by-4.0",
        anndata_version=anndata.__version__,
        data_modalities=["rna"],
        data_is_annotated=False,
        description=desc,
    )
    return HubModel(model_path, metadata=metadata, model_card=card)


def test_hub_model_init(request, save_path):
    model = prep_model()
    test_save_path = os.path.join(save_path, request.node.name)
    model.save(test_save_path, overwrite=True, save_anndata=True)

    # no metadata, no model card
    with pytest.raises(ValueError) as e:
        HubModel(test_save_path)
    assert str(e.value) == "No metadata found"

    # metadata, no model card
    hm = HubMetadata("0.17.4", "0.8.0", "SCVI")
    with pytest.raises(ValueError) as e:
        HubModel(test_save_path, metadata=hm)
    assert str(e.value) == "No model card found"

    # model card, no metadata
    hmch = HubModelCardHelper.from_dir(
        test_save_path,
        license_info="cc-by-4.0",
        anndata_version="0.8.0",
        model_parent_module="other_module",
    )
    with pytest.raises(ValueError) as e:
        HubModel(test_save_path, model_card=hmch.model_card)
    assert str(e.value) == "No metadata found"

    # successful inits
    hmo = HubModel(test_save_path, metadata=hm, model_card=hmch.model_card)
    assert hmo.metadata is hm
    assert hmo.model_card is hmch.model_card

    hmo = HubModel(test_save_path, metadata=hm, model_card=hmch)
    assert hmo.metadata is hm
    assert hmo.model_card is hmch.model_card

    hmch.model_card.save(os.path.join(test_save_path, _SCVI_HUB.MODEL_CARD_FILE_NAME))
    with open(os.path.join(test_save_path, _SCVI_HUB.METADATA_FILE_NAME), "w") as fp:
        json.dump(asdict(hm), fp, indent=4)
    hmo = HubModel(test_save_path)
    assert hmo.metadata == hm
    assert hmo.model_card.content == hmch.model_card.content

    assert hmo._local_dir == test_save_path
    assert hmo._model is None
    assert hmo._adata is None
    assert hmo._large_training_adata is None


def test_hub_model_load(request, save_path):
    model = prep_model()
    test_save_path = os.path.join(save_path, request.node.name)
    model.save(test_save_path, overwrite=True, save_anndata=True)

    hm = HubMetadata("0.17.4", "0.8.0", "SCVI")
    hmch = HubModelCardHelper.from_dir(
        test_save_path,
        license_info="cc-by-4.0",
        anndata_version="0.8.0",
        model_parent_module="other_module",
    )

    hmo = HubModel(test_save_path, metadata=hm, model_card=hmch.model_card)
    assert isinstance(hmo.model, scvi.model.SCVI)
    assert isinstance(hmo.model.module, scvi.module.VAE)
    assert np.array_equal(hmo.adata.X, model.adata.X)
    assert hmo.adata.obs.equals(model.adata.obs)
    assert hmo.adata.var.equals(model.adata.var)
    assert hmo.large_training_adata is None

    # with a custom adata
    test_save_path = os.path.join(save_path, request.node.name + "_2")
    model.save(test_save_path, overwrite=True)
    hmo = HubModel(test_save_path, metadata=hm, model_card=hmch.model_card)
    adata2 = scvi.data.synthetic_iid()
    hmo.load_model(adata=adata2)
    assert isinstance(hmo.model, scvi.model.SCVI)
    assert isinstance(hmo.model.module, scvi.module.VAE)
    assert np.array_equal(hmo.model.adata.X, adata2.X)
    assert hmo.adata is None
    assert hmo.large_training_adata is None

    # no adata
    hmo = HubModel(test_save_path, metadata=hm, model_card=hmch.model_card)
    with pytest.raises(ValueError):
        print(hmo.model)


@pytest.mark.parametrize("save_anndata", [True, False])
def test_hub_model_save(save_anndata: bool, save_path: str):
    model = prep_model()
    model_path = os.path.join(save_path, "model_save")
    model.save(model_path, save_anndata=save_anndata, overwrite=True)

    metadata = HubMetadata.from_dir(model_path, anndata_version=anndata.__version__)
    card = HubModelCardHelper.from_dir(
        model_path,
        license_info="cc-by-4.0",
        anndata_version=anndata.__version__,
        data_modalities=["rna"],
        data_is_annotated=False,
    )
    hub_model = HubModel(model_path, metadata=metadata, model_card=card)
    hub_model.save(overwrite=True)

    card_path = os.path.join(model_path, _SCVI_HUB.MODEL_CARD_FILE_NAME)
    assert os.path.exists(card_path)
    assert os.path.isfile(card_path)
    metadata_path = os.path.join(model_path, _SCVI_HUB.METADATA_FILE_NAME)
    assert os.path.exists(metadata_path)
    assert os.path.isfile(metadata_path)

    with pytest.raises(FileExistsError):
        hub_model.save(overwrite=False)

    hub_model.save(overwrite=True)


def test_prep_scvi_hub_model(save_path: str) -> HubModel:
    prep_scvi_hub_model(save_path)


@pytest.mark.private
def test_hub_model_large_training_adata(request, save_path):
    training_data_url = "https://huggingface.co/datasets/scvi-tools/DATASET-FOR-UNIT-TESTING-1/resolve/main/adata.h5ad"
    model = prep_model()
    test_save_path = os.path.join(save_path, request.node.name)
    model.save(test_save_path, overwrite=True)

    hm = HubMetadata("0.17.4", "0.8.0", "SCVI", training_data_url=training_data_url)
    hmch = HubModelCardHelper.from_dir(
        test_save_path,
        license_info="cc-by-4.0",
        anndata_version="0.8.0",
        model_parent_module="other_module",
    )

    hmo = HubModel(test_save_path, metadata=hm, model_card=hmch.model_card)
    assert isinstance(hmo.model, scvi.model.SCVI)
    assert isinstance(hmo.model.module, scvi.module.VAE)
    assert hmo.large_training_adata is hmo.model.adata
    assert hmo.adata is None


@pytest.mark.private
def test_hub_model_create_repo_hf(save_path: str):
    from huggingface_hub import delete_repo, repo_exists

    if repo_exists("scvi-tools/test-scvi-create"):
        delete_repo("scvi-tools/test-scvi-create", token=os.environ.get("HF_API_TOKEN", None))

    hub_model = prep_scvi_hub_model(save_path)
    hub_model.push_to_huggingface_hub(
        "scvi-tools/test-scvi-create",
        os.environ.get("HF_API_TOKEN", None),
        collection_name="test",
        repo_create=True,
    )
    delete_repo("scvi-tools/test-scvi-create", token=os.environ.get("HF_API_TOKEN", None))


@pytest.mark.private
def test_hub_model_push_to_hf(save_path: str):
    hub_model = prep_scvi_hub_model(save_path)
    hub_model.push_to_huggingface_hub(
        "scvi-tools/test-scvi",
        os.environ.get("HF_API_TOKEN", None),
        repo_create=False,
        collection_name="test",
    )

    hub_model = prep_scvi_no_anndata_hub_model(save_path)
    hub_model.push_to_huggingface_hub(
        "scvi-tools/test-scvi-no-anndata",
        os.environ.get("HF_API_TOKEN", None),
        repo_create=False,
        push_anndata=False,
        collection_name="test",
    )

    hub_model = prep_scvi_minified_hub_model(save_path)
    hub_model.push_to_huggingface_hub(
        "scvi-tools/test-scvi-minified",
        os.environ.get("HF_API_TOKEN", None),
        repo_create=False,
        collection_name="test",
    )


@pytest.mark.private
def test_hub_model_pull_from_hf():
    hub_model = HubModel.pull_from_huggingface_hub(repo_name="scvi-tools/test-scvi")
    assert hub_model.model is not None
    assert hub_model.adata is not None

    hub_model = HubModel.pull_from_huggingface_hub(repo_name="scvi-tools/test-scvi-minified")
    assert hub_model.model is not None
    assert hub_model.adata is not None

    hub_model = HubModel.pull_from_huggingface_hub(repo_name="scvi-tools/test-scvi-no-anndata")
    with pytest.raises(FileNotFoundError):
        _ = hub_model.model

    adata = synthetic_iid()
    hub_model.load_model(adata=adata)
    assert hub_model.model is not None
    assert hub_model.adata is None


@pytest.mark.private
def test_hub_model_push_to_s3(save_path: str):
    hub_model = prep_scvi_hub_model(save_path)
    hub_model.push_to_s3("scvi-tools", "tests/hub/test-scvi")

    hub_model = prep_scvi_no_anndata_hub_model(save_path)
    with pytest.raises(ValueError):
        hub_model.push_to_s3("scvi-tools", "tests/hub/test-scvi-no-anndata", push_anndata=True)
    hub_model.push_to_s3("scvi-tools", "tests/hub/test-scvi-no-anndata", push_anndata=False)

    hub_model = prep_scvi_minified_hub_model(save_path)
    hub_model.push_to_s3("scvi-tools", "tests/hub/test-scvi-minified")


@pytest.mark.private
def test_hub_model_pull_from_s3():
    from botocore.exceptions import ClientError

    hub_model = HubModel.pull_from_s3(
        "scvi-tools",
        "tests/hub/test-scvi",
    )
    assert hub_model.model is not None
    assert hub_model.adata is not None

    hub_model = HubModel.pull_from_s3("scvi-tools", "tests/hub/test-scvi-minified")
    assert hub_model.model is not None
    assert hub_model.adata is not None

    with pytest.raises(ClientError):
        hub_model = HubModel.pull_from_s3("scvi-tools", "tests/hub/test-scvi-no-anndata")

    hub_model = HubModel.pull_from_s3(
        "scvi-tools",
        "tests/hub/test-scvi-no-anndata",
        pull_anndata=False,
    )
    with pytest.raises(ValueError):
        _ = hub_model.model

    adata = synthetic_iid()
    hub_model.load_model(adata=adata)
    assert hub_model.model is not None
    assert hub_model.adata is None



from __future__ import annotations

import logging
import warnings
from collections.abc import Iterable as IterableClass
from functools import partial
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
import torch
from anndata import AnnData

from scvi import REGISTRY_KEYS, settings
from scvi.data import AnnDataManager, fields
from scvi.data._constants import ADATA_MINIFY_TYPE
from scvi.data._utils import _check_nonnegative_integers, _get_adata_minify_type
from scvi.dataloaders import DataSplitter
from scvi.model._utils import (
    _get_batch_code_from_category,
    _get_var_names_from_manager,
    _init_library_size,
    cite_seq_raw_counts_properties,
    get_max_epochs_heuristic,
)
from scvi.model.base._de_core import _de_core
from scvi.module import TOTALVAE
from scvi.train import AdversarialTrainingPlan, TrainRunner
from scvi.utils._docstrings import de_dsp, devices_dsp, setup_anndata_dsp

from .base import (
    ArchesMixin,
    BaseMinifiedModeModelClass,
    BaseMudataMinifiedModeModelClass,
    RNASeqMixin,
    VAEMixin,
)

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence
    from typing import Literal

    from mudata import MuData

    from scvi._types import AnnOrMuData, Number

logger = logging.getLogger(__name__)


class TOTALVI(
    RNASeqMixin,
    VAEMixin,
    ArchesMixin,
    BaseMinifiedModeModelClass,
    BaseMudataMinifiedModeModelClass,
):
    """total Variational Inference :cite:p:`GayosoSteier21`.

    Parameters
    ----------
    adata
        AnnOrMuData object that has been registered via :meth:`~scvi.model.TOTALVI.setup_anndata`
        or :meth:`~scvi.model.TOTALVI.setup_mudata`.
    n_latent
        Dimensionality of the latent space.
    gene_dispersion
        One of the following:

        * ``'gene'`` - genes_dispersion parameter of NB is constant per gene across cells
        * ``'gene-batch'`` - genes_dispersion can differ between different batches
        * ``'gene-label'`` - genes_dispersion can differ between different labels
    protein_dispersion
        One of the following:

        * ``'protein'`` - protein_dispersion parameter is constant per protein across cells
        * ``'protein-batch'`` - protein_dispersion can differ between different batches NOT TESTED
        * ``'protein-label'`` - protein_dispersion can differ between different labels NOT TESTED
    gene_likelihood
        One of:

        * ``'nb'`` - Negative binomial distribution
        * ``'zinb'`` - Zero-inflated negative binomial distribution
    latent_distribution
        One of:

        * ``'normal'`` - Normal distribution
        * ``'ln'`` - Logistic normal distribution (Normal(0, I) transformed by softmax)
    empirical_protein_background_prior
        Set the initialization of protein background prior empirically. This option fits a GMM for
        each of 100 cells per batch and averages the distributions. Note that even with this option
        set to `True`, this only initializes a parameter that is learned during inference. If
        `False`, randomly initializes. The default (`None`), sets this to `True` if greater than 10
        proteins are used.
    override_missing_proteins
        If `True`, will not treat proteins with all 0 expression in a particular batch as missing.
    **model_kwargs
        Keyword args for :class:`~scvi.module.TOTALVAE`

    Examples
    --------
    >>> mdata = mudata.read_h5mu(path_to_mudata)
    >>> scvi.model.TOTALVI.setup_mudata(
            mdata, modalities={"rna_layer": "rna", "protein_layer": "prot"}
    >>> vae = scvi.model.TOTALVI(mdata)
    >>> vae.train()
    >>> mdata.obsm["X_totalVI"] = vae.get_latent_representation()

    Notes
    -----
    See further usage examples in the following tutorials:

    1. :doc:`/tutorials/notebooks/multimodal/totalVI`
    2. :doc:`/tutorials/notebooks/multimodal/cite_scrna_integration_w_totalVI`
    3. :doc:`/tutorials/notebooks/scrna/scarches_scvi_tools`
    """

    _module_cls = TOTALVAE
    _LATENT_QZM_KEY = "totalvi_latent_qzm"
    _LATENT_QZV_KEY = "totalvi_latent_qzv"
    _data_splitter_cls = DataSplitter
    _training_plan_cls = AdversarialTrainingPlan
    _train_runner_cls = TrainRunner

    def __init__(
        self,
        adata: AnnOrMuData,
        n_latent: int = 20,
        gene_dispersion: Literal["gene", "gene-batch", "gene-label", "gene-cell"] = "gene",
        protein_dispersion: Literal["protein", "protein-batch", "protein-label"] = "protein",
        gene_likelihood: Literal["zinb", "nb"] = "nb",
        latent_distribution: Literal["normal", "ln"] = "normal",
        empirical_protein_background_prior: bool | None = None,
        override_missing_proteins: bool = False,
        **model_kwargs,
    ):
        super().__init__(adata)
        self.protein_state_registry = self.adata_manager.get_state_registry(
            REGISTRY_KEYS.PROTEIN_EXP_KEY
        )
        if (
            fields.ProteinObsmField.PROTEIN_BATCH_MASK in self.protein_state_registry
            and not override_missing_proteins
        ):
            batch_mask = self.protein_state_registry.protein_batch_mask
            msg = (
                "Some proteins have all 0 counts in some batches. "
                "These proteins will be treated as missing measurements; however, "
                "this can occur due to experimental design/biology. "
                "Reinitialize the model with `override_missing_proteins=True`,"
                "to override this behavior."
            )
            warnings.warn(msg, UserWarning, stacklevel=settings.warnings_stacklevel)
            self._use_adversarial_classifier = True
        else:
            batch_mask = None
            self._use_adversarial_classifier = False

        emp_prior = (
            empirical_protein_background_prior
            if empirical_protein_background_prior is not None
            else (self.summary_stats.n_proteins > 10)
        )
        if emp_prior and self.minified_data_type != ADATA_MINIFY_TYPE.LATENT_POSTERIOR:
            prior_mean, prior_scale = self._get_totalvi_protein_priors(adata)
        else:
            prior_mean, prior_scale = None, None

        n_cats_per_cov = (
            self.adata_manager.get_state_registry(REGISTRY_KEYS.CAT_COVS_KEY)[
                fields.CategoricalJointObsField.N_CATS_PER_KEY
            ]
            if REGISTRY_KEYS.CAT_COVS_KEY in self.adata_manager.data_registry
            else None
        )

        n_batch = self.summary_stats.n_batch
        use_size_factor_key = REGISTRY_KEYS.SIZE_FACTOR_KEY in self.adata_manager.data_registry
        library_log_means, library_log_vars = None, None
        if (
            not use_size_factor_key
            and self.minified_data_type != ADATA_MINIFY_TYPE.LATENT_POSTERIOR
        ):
            library_log_means, library_log_vars = _init_library_size(self.adata_manager, n_batch)

        self.module = self._module_cls(
            n_input_genes=self.summary_stats.n_vars,
            n_input_proteins=self.summary_stats.n_proteins,
            n_batch=n_batch,
            n_latent=n_latent,
            n_continuous_cov=self.summary_stats.get("n_extra_continuous_covs", 0),
            n_cats_per_cov=n_cats_per_cov,
            gene_dispersion=gene_dispersion,
            protein_dispersion=protein_dispersion,
            gene_likelihood=gene_likelihood,
            latent_distribution=latent_distribution,
            protein_batch_mask=batch_mask,
            protein_background_prior_mean=prior_mean,
            protein_background_prior_scale=prior_scale,
            use_size_factor_key=use_size_factor_key,
            library_log_means=library_log_means,
            library_log_vars=library_log_vars,
            **model_kwargs,
        )
        self.module.minified_data_type = self.minified_data_type
        self._model_summary_string = (
            f"TotalVI Model with the following params: \nn_latent: {n_latent}, "
            f"gene_dispersion: {gene_dispersion}, protein_dispersion: {protein_dispersion}, "
            f"gene_likelihood: {gene_likelihood}, latent_distribution: {latent_distribution}"
        )
        self.init_params_ = self._get_init_params(locals())
        if self.registry_["setup_method_name"] == "setup_mudata":
            original_dict = self.registry_["setup_args"]["modalities"]
            self.modalities = {
                "rna_layer": original_dict.get("rna_layer"),
                "protein_layer": original_dict.get("protein_layer"),
            }
        else:
            self.modalities = None

    @devices_dsp.dedent
    def train(
        self,
        max_epochs: int | None = None,
        lr: float = 4e-3,
        accelerator: str = "auto",
        devices: int | list[int] | str = "auto",
        train_size: float | None = None,
        validation_size: float | None = None,
        shuffle_set_split: bool = True,
        batch_size: int = 256,
        early_stopping: bool = True,
        check_val_every_n_epoch: int | None = None,
        reduce_lr_on_plateau: bool = True,
        n_steps_kl_warmup: int | None = None,
        n_epochs_kl_warmup: int | None = None,
        adversarial_classifier: bool | None = None,
        datasplitter_kwargs: dict | None = None,
        plan_kwargs: dict | None = None,
        external_indexing: list[np.array] = None,
        **kwargs,
    ):
        """Trains the model using amortized variational inference.

        Parameters
        ----------
        max_epochs
            Number of passes through the dataset.
        lr
            Learning rate for optimization.
        %(param_accelerator)s
        %(param_devices)s
        train_size
            Size of training set in the range [0.0, 1.0].
        validation_size
            Size of the test set. If `None`, defaults to 1 - `train_size`. If
            `train_size + validation_size < 1`, the remaining cells belong to a test set.
        shuffle_set_split
            Whether to shuffle indices before splitting. If `False`, the val, train, and test set
            are split in the sequential order of the data according to `validation_size` and
            `train_size` percentages.
        batch_size
            Minibatch size to use during training.
        early_stopping
            Whether to perform early stopping with respect to the validation set.
        check_val_every_n_epoch
            Check val every n train epochs. By default, val is not checked, unless `early_stopping`
            is `True` or `reduce_lr_on_plateau` is `True`. If either of the latter conditions are
            met, val is checked every epoch.
        reduce_lr_on_plateau
            Reduce learning rate on plateau of validation metric (default is ELBO).
        n_steps_kl_warmup
            Number of training steps (minibatches) to scale weight on KL divergences from 0 to 1.
            Only activated when `n_epochs_kl_warmup` is set to None. If `None`, defaults
            to `floor(0.75 * adata.n_obs)`.
        n_epochs_kl_warmup
            Number of epochs to scale weight on KL divergences from 0 to 1.
            Overrides `n_steps_kl_warmup` when both are not `None`.
        adversarial_classifier
            Whether to use adversarial classifier in the latent space. This helps mixing when
            there are missing proteins in any of the batches. Defaults to `True` is missing
            proteins are detected.
        datasplitter_kwargs
            Additional keyword arguments passed into :class:`~scvi.dataloaders.DataSplitter`.
        plan_kwargs
            Keyword args for :class:`~scvi.train.AdversarialTrainingPlan`. Keyword arguments passed
            to `train()` will overwrite values present in `plan_kwargs`, when appropriate.
        external_indexing
            A list of data split indices in the order of training, validation, and test sets.
            Validation and test set are not required and can be left empty.
        **kwargs
            Other keyword args for :class:`~scvi.train.Trainer`.
        """
        if adversarial_classifier is None:
            adversarial_classifier = self._use_adversarial_classifier
        n_steps_kl_warmup = (
            n_steps_kl_warmup if n_steps_kl_warmup is not None else int(0.75 * self.adata.n_obs)
        )
        if reduce_lr_on_plateau:
            check_val_every_n_epoch = 1

        update_dict = {
            "lr": lr,
            "adversarial_classifier": adversarial_classifier,
            "reduce_lr_on_plateau": reduce_lr_on_plateau,
            "n_epochs_kl_warmup": n_epochs_kl_warmup,
            "n_steps_kl_warmup": n_steps_kl_warmup,
        }
        if plan_kwargs is not None:
            plan_kwargs.update(update_dict)
        else:
            plan_kwargs = update_dict

        if max_epochs is None:
            max_epochs = get_max_epochs_heuristic(self.adata.n_obs)

        plan_kwargs = plan_kwargs if isinstance(plan_kwargs, dict) else {}
        datasplitter_kwargs = datasplitter_kwargs or {}

        data_splitter = self._data_splitter_cls(
            self.adata_manager,
            train_size=train_size,
            validation_size=validation_size,
            shuffle_set_split=shuffle_set_split,
            batch_size=batch_size,
            external_indexing=external_indexing,
            **datasplitter_kwargs,
        )
        training_plan = self._training_plan_cls(self.module, **plan_kwargs)
        runner = self._train_runner_cls(
            self,
            training_plan=training_plan,
            data_splitter=data_splitter,
            max_epochs=max_epochs,
            accelerator=accelerator,
            devices=devices,
            early_stopping=early_stopping,
            check_val_every_n_epoch=check_val_every_n_epoch,
            **kwargs,
        )
        return runner()

    @torch.inference_mode()
    def get_latent_library_size(
        self,
        adata: AnnData | None = None,
        indices: Sequence[int] | None = None,
        give_mean: bool = True,
        batch_size: int | None = None,
    ) -> np.ndarray:
        r"""Returns the latent library size for each cell.

        This is denoted as :math:`\ell_n` in the totalVI paper.

        Parameters
        ----------
        adata
            AnnData object with equivalent structure to initial AnnData. If `None`, defaults to the
            AnnData object used to initialize the model.
        indices
            Indices of cells in adata to use. If `None`, all cells are used.
        give_mean
            Return the mean or a sample from the posterior distribution.
        batch_size
            Minibatch size for data loading into model. Defaults to `scvi.settings.batch_size`.
        """
        self._check_if_trained(warn=False)

        adata = self._validate_anndata(adata)
        post = self._make_data_loader(adata=adata, indices=indices, batch_size=batch_size)

        libraries = []
        for tensors in post:
            inference_inputs = self.module._get_inference_input(tensors)
            outputs = self.module.inference(**inference_inputs)
            if give_mean:
                ql = outputs["ql"]
                library = torch.exp(ql.loc + 0.5 * (ql.scale**2))
            else:
                library = outputs["library_gene"]
            libraries += [library.cpu()]
        return torch.cat(libraries).numpy()

    @torch.inference_mode()
    def get_normalized_expression(
        self,
        adata=None,
        indices=None,
        n_samples_overall: int | None = None,
        transform_batch: Sequence[Number | str] | None = None,
        gene_list: Sequence[str] | None = None,
        protein_list: Sequence[str] | None = None,
        library_size: float | Literal["latent"] | None = 1,
        n_samples: int = 1,
        sample_protein_mixing: bool = False,
        scale_protein: bool = False,
        include_protein_background: bool = False,
        batch_size: int | None = None,
        return_mean: bool = True,
        return_numpy: bool | None = None,
    ) -> tuple[np.ndarray | pd.DataFrame, np.ndarray | pd.DataFrame]:
        r"""Returns the normalized gene expression and protein expression.

        This is denoted as :math:`\rho_n` in the totalVI paper for genes, and TODO
        for proteins, :math:`(1-\pi_{nt})\alpha_{nt}\beta_{nt}`.

        Parameters
        ----------
        adata
            AnnData object with equivalent structure to initial AnnData. If `None`, defaults to the
            AnnData object used to initialize the model.
        indices
            Indices of cells in adata to use. If `None`, all cells are used.
        n_samples_overall
            Number of samples to use in total
        transform_batch
            Batch to condition on.
            If transform_batch is:

            - None, then real observed batch is used
            - int, then batch transform_batch is used
            - List[int], then average over batches in list
        gene_list
            Return frequencies of expression for a subset of genes.
            This can save memory when working with large datasets and few genes are
            of interest.
        protein_list
            Return protein expression for a subset of genes.
            This can save memory when working with large datasets and few genes are
            of interest.
        library_size
            Scale the expression frequencies to a common library size.
            This allows gene expression levels to be interpreted on a common scale of relevant
            magnitude.
        n_samples
            Get sample scale from multiple samples.
        sample_protein_mixing
            Sample mixing bernoulli, setting background to zero
        scale_protein
            Make protein expression sum to 1
        include_protein_background
            Include background component for protein expression
        batch_size
            Minibatch size for data loading into model. Defaults to `scvi.settings.batch_size`.
        return_mean
            Whether to return the mean of the samples.
        return_numpy
            Return a `np.ndarray` instead of a `pd.DataFrame`. Includes gene
            names as columns. If either n_samples=1 or return_mean=True, defaults to False.
            Otherwise, it defaults to True.

        Returns
        -------
        - **gene_normalized_expression** - normalized expression for RNA
        - **protein_normalized_expression** - normalized expression for proteins

        If ``n_samples`` > 1 and ``return_mean`` is False, then the shape is
        ``(samples, cells, genes)``. Otherwise, shape is ``(cells, genes)``. Return type is
        ``pd.DataFrame`` unless ``return_numpy`` is True.
        """
        adata = self._validate_anndata(adata)
        adata_manager = self.get_anndata_manager(adata)
        if indices is None:
            indices = np.arange(adata.n_obs)
        if n_samples_overall is not None:
            indices = np.random.choice(indices, n_samples_overall)
        post = self._make_data_loader(adata=adata, indices=indices, batch_size=batch_size)

        if gene_list is None:
            gene_mask = slice(None)
        else:
            all_genes = _get_var_names_from_manager(adata_manager)
            gene_mask = [True if gene in gene_list else False for gene in all_genes]
        if protein_list is None:
            protein_mask = slice(None)
        else:
            all_proteins = self.protein_state_registry.column_names
            protein_mask = [True if p in protein_list else False for p in all_proteins]
        if indices is None:
            indices = np.arange(adata.n_obs)

        if n_samples > 1 and return_mean is False:
            if return_numpy is False:
                warnings.warn(
                    "`return_numpy` must be `True` if `n_samples > 1` and `return_mean` "
                    "is `False`, returning an `np.ndarray`.",
                    UserWarning,
                    stacklevel=settings.warnings_stacklevel,
                )
            return_numpy = True

        if not isinstance(transform_batch, IterableClass):
            transform_batch = [transform_batch]

        transform_batch = _get_batch_code_from_category(adata_manager, transform_batch)

        scale_list_gene = []
        scale_list_pro = []

        for tensors in post:
            x = tensors[REGISTRY_KEYS.X_KEY]
            y = tensors[REGISTRY_KEYS.PROTEIN_EXP_KEY]
            px_scale = torch.zeros_like(x)[..., gene_mask]
            py_scale = torch.zeros_like(y)[..., protein_mask]
            if n_samples > 1:
                px_scale = torch.stack(n_samples * [px_scale])
                py_scale = torch.stack(n_samples * [py_scale])
            for b in transform_batch:
                generative_kwargs = {"transform_batch": b}
                inference_kwargs = {"n_samples": n_samples}
                _, generative_outputs = self.module.forward(
                    tensors=tensors,
                    inference_kwargs=inference_kwargs,
                    generative_kwargs=generative_kwargs,
                    compute_loss=False,
                )
                if library_size == "latent":
                    px_scale += generative_outputs["px_"]["rate"].cpu()[..., gene_mask]
                else:
                    px_scale += generative_outputs["px_"]["scale"].cpu()[..., gene_mask]

                py_ = generative_outputs["py_"]
                # probability of background
                protein_mixing = 1 / (1 + torch.exp(-py_["mixing"].cpu()))
                if sample_protein_mixing is True:
                    protein_mixing = torch.distributions.Bernoulli(protein_mixing).sample()
                protein_val = py_["rate_fore"].cpu() * (1 - protein_mixing)
                if include_protein_background is True:
                    protein_val += py_["rate_back"].cpu() * protein_mixing

                if scale_protein is True:
                    protein_val = torch.nn.functional.normalize(protein_val, p=1, dim=-1)
                protein_val = protein_val[..., protein_mask]
                py_scale += protein_val
            px_scale /= len(transform_batch)
            py_scale /= len(transform_batch)
            scale_list_gene.append(px_scale)
            scale_list_pro.append(py_scale)

        if n_samples > 1:
            # concatenate along batch dimension -> result shape = (samples, cells, features)
            scale_list_gene = torch.cat(scale_list_gene, dim=1)
            scale_list_pro = torch.cat(scale_list_pro, dim=1)
            # (cells, features, samples)
            scale_list_gene = scale_list_gene.permute(1, 2, 0)
            scale_list_pro = scale_list_pro.permute(1, 2, 0)
        else:
            scale_list_gene = torch.cat(scale_list_gene, dim=0)
            scale_list_pro = torch.cat(scale_list_pro, dim=0)

        if return_mean is True and n_samples > 1:
            scale_list_gene = torch.mean(scale_list_gene, dim=-1)
            scale_list_pro = torch.mean(scale_list_pro, dim=-1)

        scale_list_gene = scale_list_gene.cpu().numpy()
        scale_list_pro = scale_list_pro.cpu().numpy()
        if return_numpy is None or return_numpy is False:
            gene_df = pd.DataFrame(
                scale_list_gene,
                columns=_get_var_names_from_manager(adata_manager)[gene_mask],
                index=adata.obs_names[indices],
            )
            protein_names = self.protein_state_registry.column_names
            pro_df = pd.DataFrame(
                scale_list_pro,
                columns=protein_names[protein_mask],
                index=adata.obs_names[indices],
            )

            return gene_df, pro_df
        else:
            return scale_list_gene, scale_list_pro

    @torch.inference_mode()
    def get_protein_foreground_probability(
        self,
        adata: AnnData | None = None,
        indices: Sequence[int] | None = None,
        transform_batch: Sequence[Number | str] | None = None,
        protein_list: Sequence[str] | None = None,
        n_samples: int = 1,
        batch_size: int | None = None,
        return_mean: bool = True,
        return_numpy: bool | None = None,
    ):
        r"""Returns the foreground probability for proteins.

        This is denoted as :math:`(1 - \pi_{nt})` in the totalVI paper.

        Parameters
        ----------
        adata
            AnnData object with equivalent structure to initial AnnData. If `None`, defaults to the
            AnnData object used to initialize the model.
        indices
            Indices of cells in adata to use. If `None`, all cells are used.
        transform_batch
            Batch to condition on.
            If transform_batch is:

            - None, then real observed batch is used
            - int, then batch transform_batch is used
            - List[int], then average over batches in list
        protein_list
            Return protein expression for a subset of genes.
            This can save memory when working with large datasets and few genes are
            of interest.
        n_samples
            Number of posterior samples to use for estimation.
        batch_size
            Minibatch size for data loading into model. Defaults to `scvi.settings.batch_size`.
        return_mean
            Whether to return the mean of the samples.
        return_numpy
            Return a :class:`~numpy.ndarray` instead of a :class:`~pandas.DataFrame`. DataFrame
            includes gene names as columns. If either `n_samples=1` or `return_mean=True`, defaults
            to `False`. Otherwise, it defaults to `True`.

        Returns
        -------
        - **foreground_probability** - probability foreground for each protein

        If `n_samples` > 1 and `return_mean` is False, then the shape is `(samples, cells, genes)`.
        Otherwise, shape is `(cells, genes)`. In this case, return type is
        :class:`~pandas.DataFrame` unless `return_numpy` is True.
        """
        adata = self._validate_anndata(adata)
        post = self._make_data_loader(adata=adata, indices=indices, batch_size=batch_size)

        if protein_list is None:
            protein_mask = slice(None)
        else:
            all_proteins = self.protein_state_registry.column_names
            protein_mask = [True if p in protein_list else False for p in all_proteins]

        if n_samples > 1 and return_mean is False:
            if return_numpy is False:
                warnings.warn(
                    "`return_numpy` must be `True` if `n_samples > 1` and `return_mean` "
                    "is `False`, returning an `np.ndarray`.",
                    UserWarning,
                    stacklevel=settings.warnings_stacklevel,
                )
            return_numpy = True
        if indices is None:
            indices = np.arange(adata.n_obs)

        py_mixings = []
        if not isinstance(transform_batch, IterableClass):
            transform_batch = [transform_batch]

        transform_batch = _get_batch_code_from_category(self.adata_manager, transform_batch)
        for tensors in post:
            y = tensors[REGISTRY_KEYS.PROTEIN_EXP_KEY]
            py_mixing = torch.zeros_like(y[..., protein_mask])
            if n_samples > 1:
                py_mixing = torch.stack(n_samples * [py_mixing])
            for b in transform_batch:
                generative_kwargs = {"transform_batch": b}
                inference_kwargs = {"n_samples": n_samples}
                _, generative_outputs = self.module.forward(
                    tensors=tensors,
                    inference_kwargs=inference_kwargs,
                    generative_kwargs=generative_kwargs,
                    compute_loss=False,
                )
                py_mixing += torch.sigmoid(generative_outputs["py_"]["mixing"])[
                    ..., protein_mask
                ].cpu()
            py_mixing /= len(transform_batch)
            py_mixings += [py_mixing]
        if n_samples > 1:
            # concatenate along batch dimension -> result shape = (samples, cells, features)
            py_mixings = torch.cat(py_mixings, dim=1)
            # (cells, features, samples)
            py_mixings = py_mixings.permute(1, 2, 0)
        else:
            py_mixings = torch.cat(py_mixings, dim=0)

        if return_mean is True and n_samples > 1:
            py_mixings = torch.mean(py_mixings, dim=-1)

        py_mixings = py_mixings.cpu().numpy()

        if return_numpy is True:
            return 1 - py_mixings
        else:
            pro_names = self.protein_state_registry.column_names
            foreground_prob = pd.DataFrame(
                1 - py_mixings,
                columns=pro_names[protein_mask],
                index=adata.obs_names[indices],
            )
            return foreground_prob

    def _expression_for_de(
        self,
        adata=None,
        indices=None,
        n_samples_overall=None,
        transform_batch: Sequence[Number | str] | None = None,
        scale_protein=False,
        batch_size: int | None = None,
        sample_protein_mixing=False,
        include_protein_background=False,
        protein_prior_count=0.5,
    ):
        rna, protein = self.get_normalized_expression(
            adata=adata,
            indices=indices,
            n_samples_overall=n_samples_overall,
            transform_batch=transform_batch,
            return_numpy=True,
            n_samples=1,
            batch_size=batch_size,
            scale_protein=scale_protein,
            sample_protein_mixing=sample_protein_mixing,
            include_protein_background=include_protein_background,
        )
        protein += protein_prior_count

        joint = np.concatenate([rna, protein], axis=1)
        return joint

    @de_dsp.dedent
    def differential_expression(
        self,
        adata: AnnData | None = None,
        groupby: str | None = None,
        group1: Iterable[str] | None = None,
        group2: str | None = None,
        idx1: Sequence[int] | Sequence[bool] | str | None = None,
        idx2: Sequence[int] | Sequence[bool] | str | None = None,
        mode: Literal["vanilla", "change"] = "change",
        delta: float = 0.25,
        batch_size: int | None = None,
        all_stats: bool = True,
        batch_correction: bool = False,
        batchid1: Iterable[str] | None = None,
        batchid2: Iterable[str] | None = None,
        fdr_target: float = 0.05,
        silent: bool = False,
        protein_prior_count: float = 0.1,
        scale_protein: bool = False,
        sample_protein_mixing: bool = False,
        include_protein_background: bool = False,
        **kwargs,
    ) -> pd.DataFrame:
        r"""A unified method for differential expression analysis.

        Implements `"vanilla"` DE :cite:p:`Lopez18`. and `"change"` mode DE :cite:p:`Boyeau19`.

        Parameters
        ----------
        %(de_adata)s
        %(de_groupby)s
        %(de_group1)s
        %(de_group2)s
        %(de_idx1)s
        %(de_idx2)s
        %(de_mode)s
        %(de_delta)s
        %(de_batch_size)s
        %(de_all_stats)s
        %(de_batch_correction)s
        %(de_batchid1)s
        %(de_batchid2)s
        %(de_fdr_target)s
        %(de_silent)s
        protein_prior_count
            Prior count added to protein expression before LFC computation
        scale_protein
            Force protein values to sum to one in every single cell (post-hoc normalization)
        sample_protein_mixing
            Sample the protein mixture component, i.e., use the parameter to sample a Bernoulli
            that determines if expression is from foreground/background.
        include_protein_background
            Include the protein background component as part of the protein expression
        **kwargs
            Keyword args for :meth:`scvi.model.base.DifferentialComputation.get_bayes_factors`

        Returns
        -------
        Differential expression DataFrame.
        """
        adata = self._validate_anndata(adata)
        adata_manager = self.get_anndata_manager(adata, required=True)
        model_fn = partial(
            self._expression_for_de,
            scale_protein=scale_protein,
            sample_protein_mixing=sample_protein_mixing,
            include_protein_background=include_protein_background,
            protein_prior_count=protein_prior_count,
            batch_size=batch_size,
        )
        col_names = np.concatenate(
            [
                np.asarray(_get_var_names_from_manager(adata_manager)),
                self.protein_state_registry.column_names,
            ]
        )
        result = _de_core(
            adata_manager,
            model_fn,
            None,
            groupby,
            group1,
            group2,
            idx1,
            idx2,
            all_stats,
            cite_seq_raw_counts_properties,
            col_names,
            mode,
            batchid1,
            batchid2,
            delta,
            batch_correction,
            fdr_target,
            silent,
            **kwargs,
        )

        return result

    @torch.inference_mode()
    def posterior_predictive_sample(
        self,
        adata: AnnOrMuData | None = None,
        indices: Sequence[int] | None = None,
        n_samples: int = 1,
        batch_size: int | None = None,
        gene_list: Sequence[str] | None = None,
        protein_list: Sequence[str] | None = None,
    ) -> np.ndarray:
        r"""Generate observation samples from the posterior predictive distribution.

        The posterior predictive distribution is written as :math:`p(\hat{x}, \hat{y} \mid x, y)`.

        Parameters
        ----------
        adata
            AnnData object with equivalent structure to initial AnnData. If `None`, defaults to the
            AnnData object used to initialize the model.
        indices
            Indices of cells in adata to use. If `None`, all cells are used.
        n_samples
            Number of required samples for each cell
        batch_size
            Minibatch size for data loading into model. Defaults to `scvi.settings.batch_size`.
        gene_list
            Names of genes of interest
        protein_list
            Names of proteins of interest

        Returns
        -------
        x_new : :class:`~numpy.ndarray`
            tensor with shape (n_cells, n_genes, n_samples)
        """
        if self.module.gene_likelihood not in ["nb"]:
            raise ValueError("Invalid gene_likelihood")

        adata = self._validate_anndata(adata)
        adata_manager = self.get_anndata_manager(adata, required=True)
        if gene_list is None:
            gene_mask = slice(None)
        else:
            all_genes = _get_var_names_from_manager(adata_manager)
            gene_mask = [True if gene in gene_list else False for gene in all_genes]
        if protein_list is None:
            protein_mask = slice(None)
        else:
            all_proteins = self.protein_state_registry.column_names
            protein_mask = [True if p in protein_list else False for p in all_proteins]

        scdl = self._make_data_loader(adata=adata, indices=indices, batch_size=batch_size)

        rna_list = []
        protein_list = []
        for tensors in scdl:
            rna_sample, protein_sample = self.module.sample(tensors, n_samples=n_samples)
            rna_sample = rna_sample[..., gene_mask]
            protein_sample = protein_sample[..., protein_mask]

            rna_list += [rna_sample]
            protein_list += [protein_sample]
            if n_samples > 1:
                rna_list[-1] = np.transpose(rna_list[-1], (1, 2, 0))
                protein_list[-1] = np.transpose(protein_list[-1], (1, 2, 0))
        rna = np.concatenate(rna_list, axis=0)
        protein = np.concatenate(protein_list, axis=0)

        if isinstance(adata, AnnData):
            return {"rna": rna, "protein": protein}
        else:
            return {self.modalities["rna_layer"]: rna, self.modalities["protein_layer"]: protein}

    @torch.inference_mode()
    def _get_denoised_samples(
        self,
        adata=None,
        indices=None,
        n_samples: int = 25,
        batch_size: int = 64,
        rna_size_factor: int = 1000,
        transform_batch: int | None = None,
    ) -> np.ndarray:
        """Return samples from an adjusted posterior predictive.

        Parameters
        ----------
        adata
            AnnData object with equivalent structure to initial AnnData. If `None`, defaults to the
            AnnData object used to initialize the model.
        indices
            indices of `adata` to use
        n_samples
            How may samples per cell
        batch_size
            Minibatch size for data loading into model. Defaults to `scvi.settings.batch_size`.
        rna_size_factor
            size factor for RNA prior to sampling gamma distribution
        transform_batch
            int of which batch to condition on for all cells
        """
        adata = self._validate_anndata(adata)
        scdl = self._make_data_loader(adata=adata, indices=indices, batch_size=batch_size)

        scdl_list = []
        for tensors in scdl:
            x = tensors[REGISTRY_KEYS.X_KEY]
            y = tensors[REGISTRY_KEYS.PROTEIN_EXP_KEY]

            generative_kwargs = {"transform_batch": transform_batch}
            inference_kwargs = {"n_samples": n_samples}
            with torch.inference_mode():
                (
                    inference_outputs,
                    generative_outputs,
                ) = self.module.forward(
                    tensors,
                    inference_kwargs=inference_kwargs,
                    generative_kwargs=generative_kwargs,
                    compute_loss=False,
                )
            px_ = generative_outputs["px_"]
            py_ = generative_outputs["py_"]
            device = px_["r"].device

            pi = 1 / (1 + torch.exp(-py_["mixing"]))
            mixing_sample = torch.distributions.Bernoulli(pi).sample()
            protein_rate = py_["rate_fore"]
            rate = torch.cat((rna_size_factor * px_["scale"], protein_rate), dim=-1)
            if len(px_["r"].size()) == 2:
                px_dispersion = px_["r"]
            else:
                px_dispersion = torch.ones_like(x).to(device) * px_["r"]
            if len(py_["r"].size()) == 2:
                py_dispersion = py_["r"]
            else:
                py_dispersion = torch.ones_like(y).to(device) * py_["r"]

            dispersion = torch.cat((px_dispersion, py_dispersion), dim=-1)

            # This gamma is really l*w using scVI manuscript notation
            p = rate / (rate + dispersion)
            r = dispersion
            # TODO: NEED TORCH MPS FIX for 'aten::_standard_gamma'
            l_train = (
                torch.distributions.Gamma(r.to("cpu"), ((1 - p) / p).to("cpu")).sample().to("mps")
                if self.device.type == "mps"
                else torch.distributions.Gamma(r, (1 - p) / p).sample()
            )
            data = l_train.cpu().numpy()
            # make background 0
            data[:, :, x.shape[1] :] = data[:, :, x.shape[1] :] * (1 - mixing_sample).cpu().numpy()
            scdl_list += [data]

            scdl_list[-1] = np.transpose(scdl_list[-1], (1, 2, 0))

        return np.concatenate(scdl_list, axis=0)

    @torch.inference_mode()
    def get_feature_correlation_matrix(
        self,
        adata=None,
        indices=None,
        n_samples: int = 10,
        batch_size: int = 64,
        rna_size_factor: int = 1000,
        transform_batch: Sequence[Number | str] | None = None,
        correlation_type: Literal["spearman", "pearson"] = "spearman",
        log_transform: bool = False,
    ) -> pd.DataFrame:
        """Generate gene-gene correlation matrix using scvi uncertainty and expression.

        Parameters
        ----------
        adata
            AnnData object with equivalent structure to initial AnnData. If `None`, defaults to the
            AnnData object used to initialize the model.
        indices
            Indices of cells in adata to use. If `None`, all cells are used.
        n_samples
            Number of posterior samples to use for estimation.
        batch_size
            Minibatch size for data loading into model. Defaults to `scvi.settings.batch_size`.
        rna_size_factor
            size factor for RNA prior to sampling gamma distribution
        transform_batch
            Batches to condition on.
            If transform_batch is:

            - None, then real observed batch is used
            - int, then batch transform_batch is used
            - list of int, then values are averaged over provided batches.
        correlation_type
            One of "pearson", "spearman".
        log_transform
            Whether to log transform denoised values prior to correlation calculation.

        Returns
        -------
        Gene-protein-gene-protein correlation matrix
        """
        from scipy.stats import spearmanr

        adata = self._validate_anndata(adata)
        adata_manager = self.get_anndata_manager(adata, required=True)

        if not isinstance(transform_batch, IterableClass):
            transform_batch = [transform_batch]

        transform_batch = _get_batch_code_from_category(
            self.get_anndata_manager(adata, required=True), transform_batch
        )

        corr_mats = []
        for b in transform_batch:
            denoised_data = self._get_denoised_samples(
                n_samples=n_samples,
                batch_size=batch_size,
                rna_size_factor=rna_size_factor,
                transform_batch=b,
                indices=indices,
            )
            flattened = np.zeros((denoised_data.shape[0] * n_samples, denoised_data.shape[1]))
            for i in range(n_samples):
                flattened[denoised_data.shape[0] * (i) : denoised_data.shape[0] * (i + 1)] = (
                    denoised_data[:, :, i]
                )
            if log_transform is True:
                flattened[:, : self.n_genes] = np.log(flattened[:, : self.n_genes] + 1e-8)
                flattened[:, self.n_genes :] = np.log1p(flattened[:, self.n_genes :])
            if correlation_type == "pearson":
                corr_matrix = np.corrcoef(flattened, rowvar=False)
            else:
                corr_matrix, _ = spearmanr(flattened, axis=0)
            corr_mats.append(corr_matrix)

        corr_matrix = np.mean(np.stack(corr_mats), axis=0)
        var_names = _get_var_names_from_manager(adata_manager)
        names = np.concatenate(
            [
                np.asarray(var_names),
                self.protein_state_registry.column_names,
            ]
        )
        return pd.DataFrame(corr_matrix, index=names, columns=names)

    @torch.inference_mode()
    def get_likelihood_parameters(
        self,
        adata: AnnData | None = None,
        indices: Sequence[int] | None = None,
        n_samples: int | None = 1,
        give_mean: bool | None = False,
        batch_size: int | None = None,
    ) -> dict[str, np.ndarray]:
        r"""Estimates for the parameters of the likelihood :math:`p(x, y \mid z)`.

        Parameters
        ----------
        adata
            AnnData object with equivalent structure to initial AnnData. If `None`, defaults to the
            AnnData object used to initialize the model.
        indices
            Indices of cells in adata to use. If `None`, all cells are used.
        n_samples
            Number of posterior samples to use for estimation.
        give_mean
            Return expected value of parameters or a samples
        batch_size
            Minibatch size for data loading into model. Defaults to `scvi.settings.batch_size`.
        """
        raise NotImplementedError

    def _validate_anndata(self, adata: AnnData | None = None, copy_if_view: bool = True):
        adata = super()._validate_anndata(adata=adata, copy_if_view=copy_if_view)
        error_msg = (
            "Number of {} in anndata different from when setup_anndata was run. Please rerun "
            "setup_anndata."
        )
        if REGISTRY_KEYS.PROTEIN_EXP_KEY in self.adata_manager.data_registry.keys():
            pro_exp = self.get_from_registry(adata, REGISTRY_KEYS.PROTEIN_EXP_KEY)
            if self.summary_stats.n_proteins != pro_exp.shape[1]:
                raise ValueError(error_msg.format("proteins"))
            is_nonneg_int = _check_nonnegative_integers(pro_exp)
            if not is_nonneg_int:
                warnings.warn(
                    "Make sure the registered protein expression in anndata contains "
                    "unnormalized count data.",
                    UserWarning,
                    stacklevel=settings.warnings_stacklevel,
                )
        else:
            raise ValueError("No protein data found, please setup or transfer anndata")

        return adata

    def _get_totalvi_protein_priors(self, adata, n_cells=100):
        """Compute an empirical prior for protein background."""
        from sklearn.exceptions import ConvergenceWarning
        from sklearn.mixture import GaussianMixture

        with warnings.catch_warnings():
            warnings.filterwarnings("error")
            logger.info("Computing empirical prior initialization for protein background.")

            adata = self._validate_anndata(adata)
            adata_manager = self.get_anndata_manager(adata)
            pro_exp = adata_manager.get_from_registry(REGISTRY_KEYS.PROTEIN_EXP_KEY)
            pro_exp = pro_exp.to_numpy() if isinstance(pro_exp, pd.DataFrame) else pro_exp
            batch_mask = adata_manager.get_state_registry(REGISTRY_KEYS.PROTEIN_EXP_KEY).get(
                fields.ProteinObsmField.PROTEIN_BATCH_MASK
            )
            batch = adata_manager.get_from_registry(REGISTRY_KEYS.BATCH_KEY).ravel()
            cats = adata_manager.get_state_registry(REGISTRY_KEYS.BATCH_KEY)[
                fields.CategoricalObsField.CATEGORICAL_MAPPING_KEY
            ]
            codes = np.arange(len(cats))

            batch_avg_mus, batch_avg_scales = [], []
            for b in np.unique(codes):
                # can happen during online updates
                # the values of these batches will not be used
                num_in_batch = np.sum(batch == b)
                if num_in_batch == 0:
                    batch_avg_mus.append(0)
                    batch_avg_scales.append(1)
                    continue
                batch_pro_exp = pro_exp[batch == b]

                # non missing
                if batch_mask is not None:
                    batch_pro_exp = batch_pro_exp[:, batch_mask[str(b)]]
                    if batch_pro_exp.shape[1] < 5:
                        logger.debug(
                            f"Batch {b} has too few proteins to set prior, setting randomly."
                        )
                        batch_avg_mus.append(0.0)
                        batch_avg_scales.append(0.05)
                        continue

                # a batch is missing because it's in the reference but not query data
                # for scarches case, these values will be replaced by original state dict
                if batch_pro_exp.shape[0] == 0:
                    batch_avg_mus.append(0.0)
                    batch_avg_scales.append(0.05)
                    continue

                cells = np.random.choice(np.arange(batch_pro_exp.shape[0]), size=n_cells)
                batch_pro_exp = batch_pro_exp[cells]
                gmm = GaussianMixture(n_components=2)
                mus, scales = [], []
                # fit per cell GMM
                for c in batch_pro_exp:
                    try:
                        gmm.fit(np.log1p(c.reshape(-1, 1)))
                    # when cell is all 0
                    except ConvergenceWarning:
                        mus.append(0)
                        scales.append(0.05)
                        continue

                    means = gmm.means_.ravel()
                    sorted_fg_bg = np.argsort(means)
                    mu = means[sorted_fg_bg].ravel()[0]
                    covariances = gmm.covariances_[sorted_fg_bg].ravel()[0]
                    scale = np.sqrt(covariances)
                    mus.append(mu)
                    scales.append(scale)

                # average distribution over cells
                batch_avg_mu = np.mean(mus)
                batch_avg_scale = np.sqrt(np.sum(np.square(scales)) / (n_cells**2))

                batch_avg_mus.append(batch_avg_mu)
                batch_avg_scales.append(batch_avg_scale)

            # repeat prior for each protein
            batch_avg_mus = np.array(batch_avg_mus, dtype=np.float32).reshape(1, -1)
            batch_avg_scales = np.array(batch_avg_scales, dtype=np.float32).reshape(1, -1)
            batch_avg_mus = np.tile(batch_avg_mus, (pro_exp.shape[1], 1))
            batch_avg_scales = np.tile(batch_avg_scales, (pro_exp.shape[1], 1))

        return batch_avg_mus, batch_avg_scales

    @torch.inference_mode()
    def get_protein_background_mean(self, adata, indices, batch_size):
        """Get protein background mean."""
        adata = self._validate_anndata(adata)
        scdl = self._make_data_loader(adata=adata, indices=indices, batch_size=batch_size)
        background_mean = []
        for tensors in scdl:
            _, inference_outputs, _ = self.module.forward(tensors)
            b_mean = inference_outputs["py_"]["rate_back"]
            background_mean += [b_mean.cpu().numpy()]
        return np.concatenate(background_mean)

    @classmethod
    @setup_anndata_dsp.dedent
    def setup_anndata(
        cls,
        adata: AnnData,
        protein_expression_obsm_key: str,
        protein_names_uns_key: str | None = None,
        batch_key: str | None = None,
        layer: str | None = None,
        size_factor_key: str | None = None,
        categorical_covariate_keys: list[str] | None = None,
        continuous_covariate_keys: list[str] | None = None,
        **kwargs,
    ):
        """%(summary)s.

        Parameters
        ----------
        %(param_adata)s
        protein_expression_obsm_key
            key in `adata.obsm` for protein expression data.
        protein_names_uns_key
            key in `adata.uns` for protein names. If None, will use the column names of
            `adata.obsm[protein_expression_obsm_key]` if it is a DataFrame, else will assign
            sequential names to proteins.
        %(param_batch_key)s
        %(param_layer)s
        %(param_size_factor_key)s
        %(param_cat_cov_keys)s
        %(param_cont_cov_keys)s

        Returns
        -------
        %(returns)s
        """
        warnings.warn(
            "We recommend using setup_mudata for multi-modal data."
            "It does not influence model performance",
            DeprecationWarning,
            stacklevel=settings.warnings_stacklevel,
        )
        setup_method_args = cls._get_setup_method_args(**locals())
        batch_field = fields.CategoricalObsField(REGISTRY_KEYS.BATCH_KEY, batch_key)
        anndata_fields = [
            fields.LayerField(REGISTRY_KEYS.X_KEY, layer, is_count_data=True),
            fields.CategoricalObsField(
                REGISTRY_KEYS.LABELS_KEY, None
            ),  # Default labels field for compatibility with TOTALVAE
            batch_field,
            fields.NumericalObsField(
                REGISTRY_KEYS.SIZE_FACTOR_KEY, size_factor_key, required=False
            ),
            fields.CategoricalJointObsField(
                REGISTRY_KEYS.CAT_COVS_KEY, categorical_covariate_keys
            ),
            fields.NumericalJointObsField(REGISTRY_KEYS.CONT_COVS_KEY, continuous_covariate_keys),
            fields.ProteinObsmField(
                REGISTRY_KEYS.PROTEIN_EXP_KEY,
                protein_expression_obsm_key,
                use_batch_mask=True,
                batch_field=batch_field,
                colnames_uns_key=protein_names_uns_key,
                is_count_data=True,
            ),
        ]
        adata_manager = AnnDataManager(fields=anndata_fields, setup_method_args=setup_method_args)
        adata_manager.register_fields(adata, **kwargs)
        cls.register_manager(adata_manager)

    @classmethod
    @setup_anndata_dsp.dedent
    def setup_mudata(
        cls,
        mdata: MuData,
        rna_layer: str | None = None,
        protein_layer: str | None = None,
        batch_key: str | None = None,
        size_factor_key: str | None = None,
        categorical_covariate_keys: list[str] | None = None,
        continuous_covariate_keys: list[str] | None = None,
        modalities: dict[str, str] | None = None,
        **kwargs,
    ):
        """%(summary_mdata)s.

        Parameters
        ----------
        %(param_mdata)s
        rna_layer
            RNA layer key. If `None`, will use `.X` of specified modality key.
        protein_layer
            Protein layer key. If `None`, will use `.X` of specified modality key.
        %(param_batch_key)s
        %(param_size_factor_key)s
        %(param_cat_cov_keys)s
        %(param_cont_cov_keys)s
        %(param_modalities)s

        Examples
        --------
        >>> mdata = muon.read_10x_h5("pbmc_10k_protein_v3_filtered_feature_bc_matrix.h5")
        >>> scvi.model.TOTALVI.setup_mudata(
                mdata, modalities={"rna_layer": "rna", "protein_layer": "prot"}
            )
        >>> vae = scvi.model.TOTALVI(mdata)
        """
        setup_method_args = cls._get_setup_method_args(**locals())

        if modalities is None:
            raise ValueError("Modalities cannot be None.")
        modalities = cls._create_modalities_attr_dict(modalities, setup_method_args)

        batch_field = fields.MuDataCategoricalObsField(
            REGISTRY_KEYS.BATCH_KEY,
            batch_key,
            mod_key=modalities.batch_key,
        )
        mudata_fields = [
            fields.MuDataLayerField(
                REGISTRY_KEYS.X_KEY,
                rna_layer,
                mod_key=modalities.rna_layer,
                is_count_data=True,
                mod_required=True,
            ),
            fields.MuDataCategoricalObsField(
                REGISTRY_KEYS.LABELS_KEY,
                None,
                mod_key=None,
            ),  # Default labels field for compatibility with TOTALVAE
            batch_field,
            fields.MuDataNumericalObsField(
                REGISTRY_KEYS.SIZE_FACTOR_KEY,
                size_factor_key,
                mod_key=modalities.size_factor_key,
                required=False,
            ),
            fields.MuDataCategoricalJointObsField(
                REGISTRY_KEYS.CAT_COVS_KEY,
                categorical_covariate_keys,
                mod_key=modalities.categorical_covariate_keys,
            ),
            fields.MuDataNumericalJointObsField(
                REGISTRY_KEYS.CONT_COVS_KEY,
                continuous_covariate_keys,
                mod_key=modalities.continuous_covariate_keys,
            ),
            fields.MuDataProteinLayerField(
                REGISTRY_KEYS.PROTEIN_EXP_KEY,
                protein_layer,
                mod_key=modalities.protein_layer,
                use_batch_mask=True,
                batch_field=batch_field,
                is_count_data=True,
                mod_required=True,
            ),
        ]
        mdata_minify_type = _get_adata_minify_type(mdata)
        if mdata_minify_type is not None:
            mudata_fields += cls._get_fields_for_mudata_minification(mdata_minify_type)
        adata_manager = AnnDataManager(fields=mudata_fields, setup_method_args=setup_method_args)
        adata_manager.register_fields(mdata, **kwargs)
        cls.register_manager(adata_manager)



import torch


def broadcast_labels(o, n_broadcast=-1):
    """Utility for the semi-supervised setting.

    If y is defined(labelled batch) then one-hot encode the labels (no broadcasting needed)
    If y is undefined (unlabelled batch) then generate all possible labels (and broadcast other
    arguments if not None)
    """
    ys_ = torch.nn.functional.one_hot(
        torch.arange(n_broadcast, device=o.device, dtype=torch.long), n_broadcast
    )
    ys = ys_.repeat_interleave(o.size(-2), dim=0)
    if o.ndim == 2:
        new_o = o.repeat(n_broadcast, 1)
    elif o.ndim == 3:
        new_o = o.repeat(1, n_broadcast, 1)
    return ys, new_o


def masked_softmax(weights, mask, dim=-1, eps=1e-30):
    """Computes a softmax of ``weights`` along ``dim`` where ``mask is True``.

    Adds a small ``eps`` term in the numerator and denominator to avoid zero division.
    Taken from: https://discuss.pytorch.org/t/apply-mask-softmax/14212/15.
    Pytorch issue tracked at: https://github.com/pytorch/pytorch/issues/55056.
    """
    weight_exps = torch.exp(weights)
    masked_exps = weight_exps.masked_fill(mask == 0, eps)
    masked_sums = masked_exps.sum(dim, keepdim=True) + eps
    return masked_exps / masked_sums



import logging
from collections.abc import Callable
from typing import NamedTuple

import numpy as np
import torch
import torchmetrics
from torch import nn

from scvi.module.base import BaseModuleClass, LossOutput, auto_move_data

logger = logging.getLogger(__name__)


class _REGISTRY_KEYS_NT(NamedTuple):
    X_KEY: str = "X"
    BATCH_KEY: str = "batch"
    DNA_CODE_KEY: str = "dna_code"


REGISTRY_KEYS = _REGISTRY_KEYS_NT()


def _round(x):
    return int(np.round(x))


class _Linear(nn.Linear):
    """Linear layer with Keras default initalizations."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def reset_parameters(self):
        """Reset parameters."""
        torch.nn.init.kaiming_normal_(self.weight)
        if self.bias is not None:
            torch.nn.init.zeros_(self.bias)


class _GELU(nn.Module):
    """GELU unit approximated by a sigmoid, same as original."""

    def __init__(self):
        super().__init__()

    def forward(self, x: torch.Tensor):
        return torch.sigmoid(1.702 * x) * x


class _BatchNorm(nn.BatchNorm1d):
    """Batch normalization with Keras default initializations and scBasset defaults."""

    def __init__(self, *args, eps: int = 1e-3, **kwargs):
        # Keras uses 0.01 for momentum, but scBasset uses 0.1
        super().__init__(*args, eps=eps, **kwargs)


class _ConvBlock(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        pool_size: int = None,
        batch_norm: bool = True,
        dropout: float = 0.0,
        activation_fn: Callable | None = None,
        ceil_mode: bool = False,
    ):
        super().__init__()
        self.conv = nn.Conv1d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            padding="same",
            bias=False,
        )
        self.batch_norm = _BatchNorm(out_channels) if batch_norm else nn.Identity()
        self.pool = (
            nn.MaxPool1d(pool_size, padding=(pool_size - 1) // 2, ceil_mode=ceil_mode)
            if pool_size is not None
            else nn.Identity()
        )
        self.activation_fn = activation_fn if activation_fn is not None else _GELU()
        self.dropout = nn.Dropout(dropout)

    def forward(self, x: torch.Tensor):
        x = self.activation_fn(x)
        x = self.conv(x)
        x = self.batch_norm(x)
        x = self.dropout(x)
        x = self.pool(x)
        return x


class _DenseBlock(nn.Module):
    def __init__(
        self,
        in_features: int,
        out_features: int,
        batch_norm: bool = True,
        dropout: float = 0.2,
        activation_fn: Callable | None = None,
    ):
        super().__init__()
        self.dense = _Linear(in_features, out_features, bias=not batch_norm)
        # batch norm with Keras default epsilon
        self.batch_norm = _BatchNorm(out_features) if batch_norm else nn.Identity()
        self.activation_fn = activation_fn if activation_fn is not None else _GELU()
        self.dropout = nn.Dropout(dropout)

    def forward(self, x: torch.Tensor):
        x = self.activation_fn(x)
        x = self.dense(x)
        x = self.batch_norm(x)
        x = self.dropout(x)
        return x


class _StochasticReverseComplement(nn.Module):
    """Stochastically reverse complement a one hot encoded DNA sequence."""

    def __init__(self):
        super().__init__()

    def forward(self, seq_1hot: torch.Tensor):
        """Stochastically reverse complement a one hot encoded DNA sequence.

        Parameters
        ----------
        seq_1hot
            [batch_size, seq_depth, seq_length] sequence
        """
        if self.training:
            reverse_bool = np.random.uniform() > 0.5
            if reverse_bool:
                # Reverse on the 4dim DNA dimension (A->T, C->G, G->C, T->A)
                # Equivalent to reversing based on our encoding
                src_seq_1hot = torch.flip(seq_1hot, [-2])
                # Reverse the sequence
                src_seq_1hot = torch.flip(src_seq_1hot, [-1])
            else:
                src_seq_1hot = seq_1hot
            return src_seq_1hot, reverse_bool
        else:
            return seq_1hot, False


class _StochasticShift(nn.Module):
    """Stochastically shift a one hot encoded DNA sequence."""

    def __init__(self, shift_max=0, pad="uniform", **kwargs):
        super().__init__()
        self.shift_max = shift_max
        self.augment_shifts = np.arange(-self.shift_max, self.shift_max + 1)
        self.pad = pad

    def forward(self, seq_1hot: torch.Tensor):
        if self.training:
            shift_i = np.random.randint(0, len(self.augment_shifts))
            shift = self.augment_shifts[shift_i]
            if shift != 0:
                return self.shift_sequence(seq_1hot, shift)
            else:
                return seq_1hot
        else:
            return seq_1hot

    @staticmethod
    def shift_sequence(seq: torch.Tensor, shift: int, pad_value: float = 0.25):
        """Shift a sequence left or right by shift_amount.

        Parameters
        ----------
        seq
            [batch_size, seq_depth, seq_length] sequence
        shift
            signed shift value (torch.int32 or int)
        pad_value
            value to fill the padding (primitive or scalar tensor)
        """
        if len(seq.shape) != 3:
            raise ValueError("input sequence should be rank 3")

        sseq = torch.roll(seq, shift, dims=-1)
        if shift > 0:
            sseq[..., :shift] = pad_value
        else:
            sseq[..., shift:] = pad_value

        return sseq


class ScBassetModule(BaseModuleClass):
    """PyTorch implementation of ScBasset :cite:p:`Yuan2022`.

    Original implementation in Keras: https://github.com/calico/scBasset

    Parameters
    ----------
    n_cells
        Number of cells to predict region accessibility
    batch_ids
        Array of (n_cells,) with batch ids for each cell
    n_filters_init
        Number of filters for the initial conv layer
    n_repeat_blocks_tower
        Number of layers in the convolutional tower
    filters_mult
        Proportion by which the number of filters should inrease in the
        convolutional tower
    n_bottleneck_layer
        Size of the bottleneck layer
    batch_norm
        Whether to apply batch norm across model layers
    dropout
        Dropout rate across layers, by default we do not do it for
        convolutional layers but we do it for the dense layers
    l2_reg_cell_embedding
        L2 regularization for the cell embedding layer
    """

    def __init__(
        self,
        n_cells: int,
        batch_ids: np.ndarray | None = None,
        n_filters_init: int = 288,
        n_repeat_blocks_tower: int = 6,
        filters_mult: float = 1.122,
        n_filters_pre_bottleneck: int = 256,
        n_bottleneck_layer: int = 32,
        batch_norm: bool = True,
        dropout: float = 0.0,
        l2_reg_cell_embedding: float = 0.0,
    ):
        super().__init__()
        self.l2_reg_cell_embedding = l2_reg_cell_embedding
        self.cell_embedding = nn.Parameter(torch.randn(n_bottleneck_layer, n_cells))
        self.cell_bias = nn.Parameter(torch.randn(n_cells))
        if batch_ids is not None:
            self.register_buffer("batch_ids", torch.as_tensor(batch_ids).long())
            self.n_batch = len(torch.unique(batch_ids))
            self.batch_emdedding = nn.Embedding(self.n_batch, n_bottleneck_layer)
        self.stem = _ConvBlock(
            in_channels=4,
            out_channels=n_filters_init,
            kernel_size=17,
            pool_size=3,
            dropout=dropout,
            batch_norm=batch_norm,
        )

        tower_layers = []
        curr_n_filters = n_filters_init
        for i in range(n_repeat_blocks_tower):
            new_n_filters = _round(curr_n_filters * filters_mult) if i > 0 else curr_n_filters
            tower_layers.append(
                _ConvBlock(
                    in_channels=curr_n_filters,
                    out_channels=new_n_filters,
                    kernel_size=5,
                    pool_size=2,
                    dropout=dropout,
                    batch_norm=batch_norm,
                )
            )
            curr_n_filters = new_n_filters
        self.tower = nn.Sequential(*tower_layers)

        self.pre_bottleneck = _ConvBlock(
            in_channels=curr_n_filters,
            out_channels=n_filters_pre_bottleneck,
            kernel_size=1,
            dropout=dropout,
            batch_norm=batch_norm,
            pool_size=1,
        )
        # NOTE: Bottleneck here assumes that seq_len=1344 and n_repeat_blocks_tower=6
        # seq_len and tower size are fixed by the in_features shape
        self.bottleneck = _DenseBlock(
            in_features=n_filters_pre_bottleneck * 7,
            out_features=n_bottleneck_layer,
            batch_norm=True,
            dropout=0.2,
        )
        self.stochastic_rc = _StochasticReverseComplement()
        self.stochastic_shift = _StochasticShift(3)

    def _get_inference_input(self, tensors: dict[str, torch.Tensor]):
        dna_code = tensors[REGISTRY_KEYS.DNA_CODE_KEY]

        input_dict = {"dna_code": dna_code}
        return input_dict

    @auto_move_data
    def inference(self, dna_code: torch.Tensor) -> dict[str, torch.Tensor]:
        """Inference method for the model."""
        # NOTE: `seq_len` assumed to be a fixed 1344 as in the original implementation.
        # input shape: (batch_size, seq_length)
        # output shape: (batch_size, 4, seq_length)
        h = nn.functional.one_hot(dna_code, num_classes=4).permute(0, 2, 1).float()
        h, _ = self.stochastic_rc(h)
        h = self.stochastic_shift(h)
        # input shape: (batch_size, 4, seq_length)
        # output shape: (batch_size, n_filters_stem, seq_length//3)
        # `stem` contains a max_pool1d by 3. For 1344 input, now 448
        h = self.stem(h)
        # output shape: (batch_size, n_filters_tower, seq_length//(3*2**6))
        # `tower` contains 6 (default) `max_pool1d` by 2
        # for 1344 input, now 7
        h = self.tower(h)
        # output shape: (batch_size, n_filters_pre_bottleneck=1, seq_length//(3*2**6))
        # `bottleneck` is a filter k=1 conv with no pooling
        # for 1344 input, now [batch_size, 1, 7]
        h = self.pre_bottleneck(h)
        # flatten the input
        # output shape: (batch_size, n_filters_pre_bottleneck * (seq_length//(3*2**6)))
        # for 1344 input, now [batch_size, 7]
        h = h.view(h.shape[0], -1)
        # Regions by bottleneck layer dim
        # output shape: (batch_size, n_bottleneck_layer)
        h = self.bottleneck(h)
        h = _GELU()(h)
        return {"region_embedding": h}

    def _get_generative_input(
        self,
        tensors: dict[str, torch.Tensor],
        inference_outputs: dict[str, torch.Tensor],
    ):
        region_embedding = inference_outputs["region_embedding"]
        input_dict = {"region_embedding": region_embedding}

        return input_dict

    def _get_accessibility(
        self,
        dna_codes: torch.Tensor,
        batch_size: int = None,
    ) -> torch.Tensor:
        """Perform minibatch inference of accessibility scores."""
        accessibility = torch.zeros(
            size=(
                dna_codes.shape[0],
                self.cell_bias.shape[0],
            )
        )
        if batch_size is None:
            # no minibatching
            batch_size = dna_codes.shape[0]
        n_batches = accessibility.shape[0] // batch_size + int(
            (accessibility.shape[0] % batch_size) > 0
        )
        for batch in range(n_batches):
            batch_codes = dna_codes[batch * batch_size : (batch + 1) * batch_size]
            # forward passes, output is dict(region_embedding=np.ndarray: [n_seqs, n_latent=32])
            motif_rep = self.inference(dna_code=batch_codes)
            # output is dict(reconstruction_logits=np.ndarray: [n_seqs, n_cells])
            batch_acc = self.generative(region_embedding=motif_rep["region_embedding"])[
                "reconstruction_logits"
            ]
            accessibility[batch * batch_size : (batch + 1) * batch_size] = batch_acc
        return accessibility

    def generative(self, region_embedding: torch.Tensor) -> dict[str, torch.Tensor]:
        """Generative method for the model."""
        if hasattr(self, "batch_ids"):
            # embeddings dim by cells dim
            cell_batch_embedding = self.batch_emdedding(self.batch_ids).squeeze(-2).T
        else:
            cell_batch_embedding = 0
        accessibility = region_embedding @ (self.cell_embedding + cell_batch_embedding)
        accessibility += self.cell_bias
        return {"reconstruction_logits": accessibility}

    def loss(self, tensors, inference_outputs, generative_outputs) -> LossOutput:
        """Loss function for the model."""
        reconstruction_logits = generative_outputs["reconstruction_logits"]
        target = tensors[REGISTRY_KEYS.X_KEY]
        loss_fn = nn.BCEWithLogitsLoss(reduction="none")
        full_loss = loss_fn(reconstruction_logits, target)
        reconstruction_loss = full_loss.sum(dim=-1)
        loss = reconstruction_loss.sum() / (
            reconstruction_logits.shape[0] * reconstruction_logits.shape[1]
        )
        if self.l2_reg_cell_embedding > 0:
            loss += self.l2_reg_cell_embedding * torch.square(self.cell_embedding).mean()
        auroc = torchmetrics.functional.auroc(
            torch.sigmoid(reconstruction_logits).ravel(),
            target.int().ravel(),
            task="binary",
        )
        return LossOutput(
            loss=loss,
            reconstruction_loss=reconstruction_loss,
            extra_metrics={"auroc": auroc},
        )



from torch.nn import Linear

from scvi.data import synthetic_iid
from scvi.external import POISSONVI


def test_poissonvi():
    adata = synthetic_iid(batch_size=100)
    POISSONVI.setup_anndata(adata)
    model = POISSONVI(adata)
    model.train(max_epochs=1)
    model.get_latent_representation()
    model.get_accessibility_estimates()
    model.get_region_factors()


def test_poissonvi_default_params():
    from scvi.model import PEAKVI

    adata = synthetic_iid(batch_size=100)
    POISSONVI.setup_anndata(adata)
    PEAKVI.setup_anndata(adata)
    poissonvi = POISSONVI(adata)
    peakvi = PEAKVI(adata)

    assert poissonvi.module.n_latent == peakvi.module.n_latent
    assert poissonvi.module.latent_distribution == peakvi.module.latent_distribution
    poisson_encoder = poissonvi.module.z_encoder.encoder
    poisson_mean_encoder = poissonvi.module.z_encoder.mean_encoder
    poisson_decoder = poissonvi.module.decoder.px_decoder
    assert len(poisson_encoder.fc_layers) == peakvi.module.n_layers_encoder
    assert len(poisson_decoder.fc_layers) == peakvi.module.n_layers_encoder
    assert poisson_encoder.fc_layers[-1][0].in_features == peakvi.module.n_hidden
    assert poisson_decoder.fc_layers[-1][0].in_features == peakvi.module.n_hidden
    assert poisson_mean_encoder.out_features == peakvi.module.n_latent
    assert poisson_decoder.fc_layers[0][0].in_features == peakvi.module.n_latent


def test_poissonvi_non_default_params(
    n_hidden: int = 50,
    n_latent: int = 5,
    n_layers: int = 2,
    dropout_rate: float = 0.4,
    latent_distribution="ln",
):
    adata = synthetic_iid(batch_size=100)
    POISSONVI.setup_anndata(adata)
    model = POISSONVI(
        adata,
        n_hidden=n_hidden,
        n_latent=n_latent,
        n_layers=n_layers,
        dropout_rate=dropout_rate,
        latent_distribution=latent_distribution,
    )

    assert model.module.n_latent == n_latent
    assert model.module.latent_distribution == latent_distribution

    encoder = model.module.z_encoder.encoder
    assert len(encoder.fc_layers) == n_layers
    linear = encoder.fc_layers[-1][0]
    assert isinstance(linear, Linear)
    assert linear.in_features == n_hidden
    mean_encoder = model.module.z_encoder.mean_encoder
    assert isinstance(mean_encoder, Linear)
    assert mean_encoder.out_features == n_latent

    model.train(max_epochs=1)
    assert model.get_latent_representation().shape[1] == n_latent



import logging
import warnings
from copy import deepcopy

import anndata
import numpy as np
import pandas as pd
import torch
from anndata import AnnData
from mudata import MuData
from scipy.sparse import csr_matrix

from scvi import settings
from scvi._types import AnnOrMuData
from scvi.data import _constants
from scvi.data._constants import _MODEL_NAME_KEY, _SETUP_ARGS_KEY, _SETUP_METHOD_NAME
from scvi.model._utils import parse_device_args
from scvi.model.base._save_load import (
    _get_var_names,
    _initialize_model,
    _load_saved_files,
    _validate_var_names,
)
from scvi.nn import FCLayers
from scvi.utils._docstrings import devices_dsp

from ._base_model import BaseModelClass

logger = logging.getLogger(__name__)

MIN_VAR_NAME_RATIO = 0.8


class ArchesMixin:
    """Universal scArches implementation."""

    @classmethod
    @devices_dsp.dedent
    def load_query_data(
        cls,
        adata: AnnOrMuData,
        reference_model: str | BaseModelClass,
        inplace_subset_query_vars: bool = False,
        accelerator: str = "auto",
        device: int | str = "auto",
        unfrozen: bool = False,
        freeze_dropout: bool = False,
        freeze_expression: bool = True,
        freeze_decoder_first_layer: bool = True,
        freeze_batchnorm_encoder: bool = True,
        freeze_batchnorm_decoder: bool = False,
        freeze_classifier: bool = True,
    ):
        """Online update of a reference model with scArches algorithm :cite:p:`Lotfollahi21`.

        Parameters
        ----------
        adata
            AnnData organized in the same way as data used to train model.
            It is not necessary to run setup_anndata,
            as AnnData is validated against the ``registry``.
        reference_model
            Either an already instantiated model of the same class, or a path to
            saved outputs for reference model.
        inplace_subset_query_vars
            Whether to subset and rearrange query vars inplace based on vars used to
            train reference model.
        %(param_accelerator)s
        %(param_device)s
        unfrozen
            Override all other freeze options for a fully unfrozen model
        freeze_dropout
            Whether to freeze dropout during training
        freeze_expression
            Freeze neurons corersponding to expression in first layer
        freeze_decoder_first_layer
            Freeze neurons corersponding to first layer in decoder
        freeze_batchnorm_encoder
            Whether to freeze batchnorm weight and bias during training for encoder
        freeze_batchnorm_decoder
            Whether to freeze batchnorm weight and bias during training for decoder
        freeze_classifier
            Whether to freeze classifier completely. Only applies to `SCANVI`.
        """
        _, _, device = parse_device_args(
            accelerator=accelerator,
            devices=device,
            return_device="torch",
            validate_single_device=True,
        )

        attr_dict, var_names, load_state_dict = _get_loaded_data(reference_model, device=device)

        if isinstance(adata, MuData):
            for modality in adata.mod:
                if inplace_subset_query_vars:
                    logger.debug(f"Subsetting {modality} query vars to reference vars.")
                    adata[modality]._inplace_subset_var(var_names[modality])
                _validate_var_names(adata[modality], var_names[modality])

        else:
            if inplace_subset_query_vars:
                logger.debug("Subsetting query vars to reference vars.")
                adata._inplace_subset_var(var_names)
            _validate_var_names(adata, var_names)

        if inplace_subset_query_vars:
            logger.debug("Subsetting query vars to reference vars.")
            adata._inplace_subset_var(var_names)
        _validate_var_names(adata, var_names)

        registry = attr_dict.pop("registry_")
        if _MODEL_NAME_KEY in registry and registry[_MODEL_NAME_KEY] != cls.__name__:
            raise ValueError("It appears you are loading a model from a different class.")

        if _SETUP_ARGS_KEY not in registry:
            raise ValueError(
                "Saved model does not contain original setup inputs. "
                "Cannot load the original setup."
            )

        setup_method = getattr(cls, registry[_SETUP_METHOD_NAME])
        setup_method(
            adata,
            source_registry=registry,
            extend_categories=True,
            allow_missing_labels=True,
            **registry[_SETUP_ARGS_KEY],
        )

        model = _initialize_model(cls, adata, attr_dict)
        adata_manager = model.get_anndata_manager(adata, required=True)

        version_split = adata_manager.registry[_constants._SCVI_VERSION_KEY].split(".")
        if int(version_split[1]) < 8 and int(version_split[0]) == 0:
            warnings.warn(
                "Query integration should be performed using models trained with "
                "version >= 0.8",
                UserWarning,
                stacklevel=settings.warnings_stacklevel,
            )

        model.to_device(device)

        # model tweaking
        new_state_dict = model.module.state_dict()
        for key, load_ten in load_state_dict.items():
            new_ten = new_state_dict[key]
            if new_ten.size() == load_ten.size():
                continue
            # new categoricals changed size
            else:
                dim_diff = new_ten.size()[-1] - load_ten.size()[-1]
                fixed_ten = torch.cat([load_ten, new_ten[..., -dim_diff:]], dim=-1)
                load_state_dict[key] = fixed_ten

        model.module.load_state_dict(load_state_dict)
        model.module.eval()

        _set_params_online_update(
            model.module,
            unfrozen=unfrozen,
            freeze_decoder_first_layer=freeze_decoder_first_layer,
            freeze_batchnorm_encoder=freeze_batchnorm_encoder,
            freeze_batchnorm_decoder=freeze_batchnorm_decoder,
            freeze_dropout=freeze_dropout,
            freeze_expression=freeze_expression,
            freeze_classifier=freeze_classifier,
        )
        model.is_trained_ = False

        return model

    @staticmethod
    def prepare_query_anndata(
        adata: AnnData,
        reference_model: str | BaseModelClass,
        return_reference_var_names: bool = False,
        inplace: bool = True,
    ) -> AnnData | pd.Index | None:
        """Prepare data for query integration.

        This function will return a new AnnData object with padded zeros
        for missing features, as well as correctly sorted features.

        Parameters
        ----------
        adata
            AnnData organized in the same way as data used to train model.
            It is not necessary to run setup_anndata,
            as AnnData is validated against the ``registry``.
        reference_model
            Either an already instantiated model of the same class, or a path to
            saved outputs for reference model.
        return_reference_var_names
            Only load and return reference var names if True.
        inplace
            Whether to subset and rearrange query vars inplace or return new AnnData.

        Returns
        -------
        Query adata ready to use in `load_query_data` unless `return_reference_var_names`
        in which case a pd.Index of reference var names is returned.
        """
        _, var_names, _ = _get_loaded_data(reference_model, device="cpu")
        var_names = pd.Index(var_names)

        if return_reference_var_names:
            return var_names

        return _pad_and_sort_query_anndata(adata, var_names, inplace)

    @staticmethod
    def prepare_query_mudata(
        mdata: MuData,
        reference_model: str | BaseModelClass,
        return_reference_var_names: bool = False,
        inplace: bool = True,
    ) -> MuData | dict[str, pd.Index] | None:
        """Prepare multimodal dataset for query integration.

        This function will return a new MuData object such that the
        AnnData objects for individual modalities are given padded zeros
        for missing features, as well as correctly sorted features.

        Parameters
        ----------
        mdata
            MuData organized in the same way as data used to train model.
            It is not necessary to run setup_mudata,
            as MuData is validated against the ``registry``.
        reference_model
            Either an already instantiated model of the same class, or a path to
            saved outputs for reference model.
        return_reference_var_names
            Only load and return reference var names if True.
        inplace
            Whether to subset and rearrange query vars inplace or return new MuData.

        Returns
        -------
        Query mudata ready to use in `load_query_data` unless `return_reference_var_names`
        in which case a dictionary of pd.Index of reference var names is returned.
        """
        attr_dict, var_names, _ = _get_loaded_data(reference_model, device="cpu")

        for modality in var_names.keys():
            var_names[modality] = pd.Index(var_names[modality])

        if return_reference_var_names:
            return var_names

        reference_modalities_dict = attr_dict["registry_"][_SETUP_ARGS_KEY]["modalities"]

        reference_modalities = reference_modalities_dict.values()
        query_modalities = mdata.mod

        for mod in reference_modalities:
            if mod not in query_modalities:
                raise ValueError(
                    "Query MuData does not contain same modalities as reference. "
                    "Cannot load the original setup."
                )

        adata_dict = {}
        for modality in reference_modalities:
            adata_out = _pad_and_sort_query_anndata(
                adata=mdata[modality],
                reference_var_names=var_names[modality],
                inplace=inplace,
            )
            adata_dict[modality] = adata_out

        if not inplace:
            return MuData(adata_dict)


def _set_params_online_update(
    module,
    unfrozen,
    freeze_decoder_first_layer,
    freeze_batchnorm_encoder,
    freeze_batchnorm_decoder,
    freeze_dropout,
    freeze_expression,
    freeze_classifier,
):
    """Freeze parts of network for scArches."""
    # do nothing if unfrozen
    if unfrozen:
        return

    mod_inference_mode = {"encoder_z2_z1", "decoder_z1_z2"}
    mod_no_hooks_yes_grad = {"l_encoder"}
    if not freeze_classifier:
        mod_no_hooks_yes_grad.add("classifier")
    parameters_yes_grad = {"background_pro_alpha", "background_pro_log_beta"}

    def no_hook_cond(key):
        one = (not freeze_expression) and "encoder" in key
        two = (not freeze_decoder_first_layer) and "px_decoder" in key
        return one or two

    def requires_grad(key):
        mod_name = key.split(".")[0]
        # linear weights and bias that need grad
        one = "fc_layers" in key and ".0." in key and mod_name not in mod_inference_mode
        # modules that need grad
        two = mod_name in mod_no_hooks_yes_grad
        three = sum([p in key for p in parameters_yes_grad]) > 0
        # batch norm option
        four = (
            "fc_layers" in key
            and ".1." in key
            and "encoder" in key
            and (not freeze_batchnorm_encoder)
        )
        five = (
            "fc_layers" in key
            and ".1." in key
            and "decoder" in key
            and (not freeze_batchnorm_decoder)
        )
        if one or two or three or four or five:
            return True
        else:
            return False

    for key, mod in module.named_modules():
        # skip over protected modules
        if key.split(".")[0] in mod_no_hooks_yes_grad:
            continue
        if isinstance(mod, FCLayers):
            hook_first_layer = False if no_hook_cond(key) else True
            mod.set_online_update_hooks(hook_first_layer)
        if isinstance(mod, torch.nn.Dropout):
            if freeze_dropout:
                mod.p = 0
        # momentum freezes the running stats of batchnorm
        freeze_batchnorm = ("decoder" in key and freeze_batchnorm_decoder) or (
            "encoder" in key and freeze_batchnorm_encoder
        )
        if isinstance(mod, torch.nn.BatchNorm1d) and freeze_batchnorm:
            mod.momentum = 0

    for key, par in module.named_parameters():
        if requires_grad(key):
            par.requires_grad = True
        else:
            par.requires_grad = False


def _get_loaded_data(reference_model, device=None):
    if isinstance(reference_model, str):
        attr_dict, var_names, load_state_dict, _ = _load_saved_files(
            reference_model, load_adata=False, map_location=device
        )
    else:
        attr_dict = reference_model._get_user_attributes()
        attr_dict = {a[0]: a[1] for a in attr_dict if a[0][-1] == "_"}
        var_names = _get_var_names(reference_model.adata)
        load_state_dict = deepcopy(reference_model.module.state_dict())

    return attr_dict, var_names, load_state_dict


def _pad_and_sort_query_anndata(
    adata: AnnData,
    reference_var_names: pd.Index,
    inplace: bool,
):
    intersection = adata.var_names.intersection(reference_var_names)
    inter_len = len(intersection)
    if inter_len == 0:
        raise ValueError(
            "No reference var names found in query data. "
            "Please rerun with return_reference_var_names=True "
            "to see reference var names."
        )

    ratio = inter_len / len(reference_var_names)
    logger.info(f"Found {ratio * 100}% reference vars in query data.")
    if ratio < MIN_VAR_NAME_RATIO:
        warnings.warn(
            f"Query data contains less than {MIN_VAR_NAME_RATIO:.0%} of reference "
            "var names. This may result in poor performance.",
            UserWarning,
            stacklevel=settings.warnings_stacklevel,
        )
    genes_to_add = reference_var_names.difference(adata.var_names)
    needs_padding = len(genes_to_add) > 0
    if needs_padding:
        padding_mtx = csr_matrix(np.zeros((adata.n_obs, len(genes_to_add))))
        adata_padding = AnnData(
            X=padding_mtx.copy(),
            layers={layer: padding_mtx.copy() for layer in adata.layers},
        )
        adata_padding.var_names = genes_to_add
        adata_padding.obs_names = adata.obs_names
        # Concatenate object
        adata_out = anndata.concat(
            [adata, adata_padding],
            axis=1,
            join="outer",
            index_unique=None,
            merge="unique",
        )
    else:
        adata_out = adata

    # also covers the case when new adata has more var names than old
    if not reference_var_names.equals(adata_out.var_names):
        adata_out._inplace_subset_var(reference_var_names)

    if inplace:
        if adata_out is not adata:
            adata._init_as_actual(adata_out)
    else:
        return adata_out
